#helper functions to process viterbi outputs
from colorama import Fore, Back, Style
import numpy as np
def label_states(vit_out, hidden_states):
    '''
    Function to convert output from testvit.c to a single sequence of labeled hidden states

    Args:
        vit_out (text file): text file output of testvit.c
        hidden_states_in (text file): text file of "." delimited states in correct order to label with

    Returns:
        labeled_seq (list): a list corresponding to the input sequence in terms of hidden_states
        pointers (dict): a dictionary of "pointers" to the beginning of a state set (P,R, or S) so we can jump there later
        MLE (float): Maximum likelihood estimate of the viterbi path

    '''
    labeled_seq = []
    #pointers to record indeces of the start of each part of our model in the read sequence
    pointers = {"P":False, "R":False, "S":False, "G2":False, "D":[], "I":[]}
    #get likelihoods from viterbi output
    MLE = float(vit_out.split("\n")[0].rstrip().split(" ")[6])
    seq = vit_out.split("\n")[3]
    seq_list = seq.rstrip().split(" ")
    seq_list = [int(i) for i in seq_list]

    #build dictionary to convert from numeric to states
    conversion = {}
    for i,state in enumerate(hidden_states):
        conversion[i+1] = state #viterbi labels are one indexed
    #convert labeled_seq
    for i,obs in enumerate(seq_list):
        labeled_seq.append(conversion[seq_list[i]])
        #check if we need to add a "pointer"
        if  "P" in conversion[seq_list[i]] and pointers["P"] == False:
            pointers["P"] = i
        if "R" in conversion[seq_list[i]] and pointers["R"] == False:
            pointers["R"] = i
        if  "S" in conversion[seq_list[i]] and pointers["S"] == False:
            pointers["S"] = i
        if conversion[seq_list[i]] == "G2" and pointers["G2"] == False:
            pointers["G2"] = i
        if "D" in conversion[seq_list[i]]: #get indeces of all deletions
            pointers["D"].append(i)
        if "I" in conversion[seq_list[i]]: #get indeces of all deletions
            pointers["I"].append(i)
    return labeled_seq, pointers,MLE


def calc_likelihood(vit_out, pointers, labeled_seq,hidden_states, read,subset_start, subset_end):
    '''
    Function to calculate likelihoods corresponding to subseqeunces to be used later

    Args:
        vit_out (text file): text file output of testvit.c
        pointers (dict): dictionary of indeces in labeled_seq corresponding to beginning of different sections of the model
        labeled_seq (str): labeled state seqeunce corresponding to viterbi output
        hidden_states (text file): text file of "." delimited states in correct order to label with
        read (str): raw read sequence
        subset_start (int): integer indicating subset start position in original read
        subset_end (int): integer indeicating subset end position in original read

    Returns:
        final (float): final likelihood corresponding to prefix --> suffix viterbi path
        final_labels(str): seqeunce labels for prefix --> suffix, "-" delimited
        repeats (str): sequence including deletion labels corresponding to only the target repeat in the seqeunce
        context (str): sequence including deletion labels corresponding to the path through prefix-->suffix
        final_repeat_like (float): likelihood of the tandem repeat sequence
        repeat_start (int): repeat start position relative to the original read
        repeat_end (int): repeat end position relative to the original read

    '''
    #we can calculate the likelihood using the labeled sequence
    likelihoods_int = vit_out.split("\n")[4]
    likelihoods_dec = vit_out.split("\n")[5]
    #process likelihood outputs, combine leading integers with decimals
    likelihood_int_list = likelihoods_int.rstrip().split(" ")
    likelihood_dec_list = likelihoods_dec.rstrip().split(" ")
    likelihood_int_list = [int(i) for i in likelihood_int_list]
    likelihood_dec_list = [abs(int(i)) for i in likelihood_dec_list]
    likelihood_list = [float('.'.join(str(i) for i in x)) for x in list(zip(likelihood_int_list,likelihood_dec_list))]
    #get the last G1 and S likelihoods
    Sn = pointers["G2"] - 1
    Gn = pointers["P"] - 1
    final = likelihood_list[Sn] - likelihood_list[Gn]
    final_repeat_like = likelihood_list[pointers["S"]-1] - likelihood_list[pointers["R"]-1] #likelihood of just the repeat

    #get repeat coordinates per read for downstream motif analysis
    #FIXME this doesn't account for the fact that deletions are included in pointers and NOT subset start should be as follows (update install)
    repeat_start = pointers["R"] + subset_start - sum(np.where(np.array(pointers["D"]) < pointers["R"], True, False))
    repeat_end = pointers["S"] + subset_start - sum(np.where(np.array(pointers["D"]) < pointers["S"], True, False))
    # repeat_end =pointers["S"]-1 + subset_start
    # repeat_start = pointers["R"] + subset_start

    final_labels = "-".join(labeled_seq[pointers["P"]:pointers["G2"]])
    num_pd = final_labels.count('PD')
    sub_read = read[pointers["P"]+num_pd:] #need to adjust for deletions but I think it maybe postion dependent

    #get indeces in labelled with deletions
    deletions = [i - pointers["P"] for i in pointers["D"]] #convert to be relative to prefix start position

    #for each deletion, add a "-" in the raw sequence (this may not be the most efficient way to do this)
    context = ""
    offset = 0
    #print(final_labels)
    for i in range(len(labeled_seq)):
        if 'G' in labeled_seq[i]:
            continue
        if 'D' in labeled_seq[i]:
            context= context + "-"
            offset+=1
        else:
            context = context + read[i-offset]

    #now our read is the same length as the labelled seq so we can index with the pointers
    R_start = pointers["R"] - pointers["P"]
    R_end = pointers["S"] - pointers["P"]
    repeats = context[R_start:R_end] #need to see if this includes last index of repeat
    return final, final_labels, repeats, context, final_repeat_like, repeat_start, repeat_end
def count_repeats(labeled_seq, pointers,repeat_len,seq):
    '''
    Function to count the number of repeats in a sequence given a sequence labeled by its hidden states

    Args:
        labeled_seq (list): hidden state annotation of query sequence
        pointers (dict): dictionary of "pointers" to the first match to prefix, repeat, and suffix
        repeat_len (int): length of target repeat
        seq (str): read sequence

    Returns:
        count (int): the number of repeats identified in the query sequence

    '''
    repeat_region = labeled_seq[pointers["R"]:pointers["S"]]
    adjusted_length = len(repeat_region)
    #count insertions to account for them in count
    for state in repeat_region:
        if state[1] == "I":
            adjusted_length -= 1
    #calculate counts from adjusted length, deletions already accounted for in labeled sequence
    count = adjusted_length/repeat_len
    return count

def print_labelled(read_id,strand,sub_labels,context,pointers,out):
    '''
    Function to output color-coded context seqeunce

    Args:
        read_id(str): read id for input sequence
        strand(str): strand of read to output
        sub_labels(str): "-" delimited labels for prefix-->suffix
        context(str): context sequence including deletion labels
        pointers(dict): dictionary of pointers corresponding to indeces of insertions, deletions, and start positions
        out(str): output file suffix
    Returns:
    None, outputs new context sequence string to context file for given target
    '''
    # FIXME there is currently an edge case where if there is a deletion at the end of the repeat the sequence will continue to be labelled in white
    R_start = pointers["R"] - pointers["P"]
    R_end = pointers["S"] - pointers["P"]
    context_list = list(context)
    I = [i - pointers["P"] for i in pointers["I"]]
    D = [i - pointers["P"] for i in pointers["D"]]
    for i in I:
        if i < R_start or i > R_end:
            context_list[i] = '\x1b[5;37;42m' + context_list[i] + Style.RESET_ALL + '\x1b[1;30;40m'
        else:
            context_list[i] = '\x1b[5;37;42m' + context_list[i] +  Style.RESET_ALL + '\x1b[1;37;40m'
    for i in D:
        if i < R_start or i > R_end:
            context_list[i] = '\x1b[5;31;41m'+" " + Style.RESET_ALL + '\x1b[1;30;40m'
        else:
            context_list[i] = '\x1b[5;31;41m' + " "+  Style.RESET_ALL + '\x1b[1;37;40m'
    #replace all inserted bases with color + base to be printed
    file = open(out,"a")
    file.write(read_id + " " + strand + " " + '\x1b[1;30;40m' + "".join(context_list[:R_start]) + Style.RESET_ALL + '\x1b[1;37;40m' + "".join(context_list[R_start:R_end]) + Style.RESET_ALL +'\x1b[1;30;40m'+ "".join(context_list[R_end:])+ Style.RESET_ALL + "\n")
    file.close()
    return
    


#I dont need all of these, remove all unused at the end, many will be encapsulated by other files
#from collections import Counter, defaultdict
import argparse
import os, glob
import gzip
#import numpy as np
import pandas as pd
from time import perf_counter
import multiprocessing as mp
import pickle as pkl

#custom imports
from profile_HMM.profile_HMM import ProfileHMM
from HMMSTR_utils.HMMSTR_utils import *
from process_read.process_read import Process_Read
from GMM_stats.GMM_stats import GMMStats

def convert_to_fasta(tsv,out):
    '''
    Function to write "read" entry for a given target for both prefix and suffix
    fasta files to be used as queries in blastn

    Args:
        tsv (Pandas Dataframe): input tsv containing name, prefix, repeat, suffix
        out (str): prefix/path for outputted files

    Returns:
        nothing, writes prefix and suffix fasta files to out_prefix.fa and out_suffix.fa respectively

    '''
    for row in tsv.itertuples():
        curr_out = open(out+"_prefix.fa",'a')
        curr_out.write(">"+row.name+"\n")
        curr_out.write(row.prefix + "\n")
        curr_out.close()
        curr_out = open(out +"_suffix.fa",'a')
        curr_out.write(">"+row.name+"\n")
        curr_out.write(row.suffix + "\n")
        curr_out.close()
    return
def build_all(curr_target, out, background,alphabet,transitions,mismatch_probs, repeat_probs,flanking_size):
    '''
    Function to initialize a ProfileHMM object for a row in input targets tsv

    Parameters
    ----------------------------------------------------------------------------
    curr_target: pandas Series
        Row from input tsv corresponding to a single target
    alphabet: list
        list of characters used to define alphabet of HMM
    background: dictionary
        Dictionary of background probabilities to use as background frequencies
    out: str
        output prefix

    Returns
    ------------------------------------------------------------------------------
    None, writes ProfileHMM in format needed for modified umdhmm-v1.02
    '''
    name = curr_target[0]
    prefix = curr_target[1].rstrip()
    repeat = curr_target[2].rstrip()
    suffix = curr_target[3].rstrip()
    #initialize ProfileHMM object
    curr_hmm = ProfileHMM(name=name,alphabet=alphabet, prefix=prefix, repeat=repeat, suffix=suffix, background=background, out=out, transitions=transitions,mismatch_probs=mismatch_probs, repeat_probs=repeat_probs,flanking_size=flanking_size)
    curr_hmm.build_profile("forward") #builds and writes forward ProfileHMM object
    curr_hmm.build_profile("reverse") #builds and writes reverse ProfileHMM object

    #write out the transition and emission matrices to pickle files for original scaling protocol CAN BE REMOVED IN FINAL
    # with open(out + "_" + name + "_A.pkl", "wb") as outfile:
    #     pkl.dump(curr_hmm.A, outfile)
    # with open(out + "_" + name + "_E.pkl", "wb") as outfile:
    #     pkl.dump(curr_hmm.E, outfile)
    return
def process_read(header,seq,hmm_file,rev_hmm_file,hidden_states,hidden_states_rev,out,targets, build_pre, mode, cutoff, k, w, use_full_seq, flanking_size):
    '''
    Wrapper function to call all Process_Read methods. This will initialize all operations
    for a given read including (1) alignment (2) target identification and filtering 
    (3) repeat identification with viterbi (4) repeat counting and (5) recording relevant metrics

    Parameters
    ----------------------------------------------------------------------------------------------
    header: str
        header for read from fasta file
    seq: str
        raw read sequence
    hmm_file: str
        path to file containing written profile hmm from build_all
    rev_hmm_file: str
        path to file continaing the reverse strand profile hmm from build_all
    hidden_states: str
        path to hidden states file 
    hidden_states_rev: str
        path to hidden states file for reverse strand
    out: str
        Output prefix for run including full or relative path
    targets: pandas DataFrame
        DataFrame containing input information for each target to search for in reads
    build_pre: str
        Path and prefix for hmm files, used in case targets have already been used and user doesn't want duplicate files
    mode: str
        Sequencing technology used, map-ont for Nanopore, pb for PacBio
    cutoff: int
        Mapq cutoff for mappy aligner to determine read maping location
    
    Returns
    ------------------------------------------------------------------------------------------------
    None, all outputs written to out
    '''
    curr_read = Process_Read(header, seq, cutoff, mode, out, k,w, use_full_seq)
    curr_read.align_mappy() #sets prefix_df and suffix_df fields for read
    # if len(curr_read.suffix_df.name) > 0:
    #     print(curr_read.prefix_df)
    #     print(curr_read.suffix_df) 
    curr_read.assign_targets(targets) #sets target_info for current read, this is a dictionary of keys=target name and values=info, all information needed for the target
    #The check for no assigned targets is in run_vit
    targets_found = curr_read.run_viterbi(hmm_file=hmm_file,rev_hmm_file=rev_hmm_file,hidden_states=hidden_states, rev_states=hidden_states_rev, out=out,build_pre=build_pre, prefix_idx = flanking_size)
    return
def ratio_gmm_stats(row, out, out_count_name,plot_hists,max_peaks, filter_outliers, bootstrap, CI_width, resample_size, allele_specific_plots,allele_specif_CIs): 
    '''
    This is a wrapper function to run stats methods on all targets. This version corresponds to our original stats methods
    which includes normalizing by a perfect match path through the hmm followed by GMM clustering and allele calls

    Args:
        target(pandas series): row of target dataframe
        out(str): output suffix
        out_count_name(str): output suffix for final result tsv

    Returns:
        None, results written out to tsv per target

    '''
    #check existence of count file, may not exist if no reads identified for a given target
    name = row[0]
    out_count_file = out + "_" + name + out_count_name
    #print(out_count_file)
    if os.path.exists(out_count_file) == False:
        print(out_count_file + " does not exist, writing null row for", name,"...")
        # returning null at the beginning of the run may be causing issues with apply, try returning a null row of the right dimensions
        curr_dict = {"name":name}
        for i in range(1,max_peaks+1):
            #curr_mean = "H"+str(i)+":mean"
            curr_median = "H"+str(i)+":median"
            curr_mode = "H"+str(i)+":mode"
            curr_sd = "H" + str(i) + ":SD"
            curr_support = "H" + str(i) + ":supporting_reads"
            #if curr_mean not in curr_dict.keys():
             #   curr_dict[curr_mean] = 0
            if curr_median not in curr_dict.keys():
                curr_dict[curr_median] = 0
            if curr_mode not in curr_dict.keys():
                curr_dict[curr_mode] = 0
            if curr_sd not in curr_dict.keys():
                curr_dict[curr_sd] = 0
            if curr_support not in curr_dict.keys():
                curr_dict[curr_support] = 0
            #curr_dict["H"+str(i)] = -1
            curr_dict["num_supporting_reads"] = 0

            #check if we were suppose to bootstrap for this run so we can return the correct number of columns
            if bootstrap:
                curr_dict["H"+str(i)+":median_CI"] = None
            if allele_specif_CIs:
                curr_dict["H"+ str(i)+":median_CI_allele_specific"] = None
        return pd.Series(curr_dict)
    #initialize GMMStats object
    gmm_stats = GMMStats(target_row=row) #contains all target attributes as well as E and A dictionaries
    final_data = gmm_stats.get_stats(out_count_file, out,plot_hists, filter_outliers)

    final_data.to_csv(out +"_"+ gmm_stats.name +"_read_info.tsv", index = False, sep="\t") #previously final_out.tsv

    #peak calling
    final_data = final_data[final_data.outlier == False].copy()
    curr_row, final_data['cluster_assignments'] = gmm_stats.call_peaks(final_data, out, max_peaks, plot=plot_hists, save_allele_plots = allele_specific_plots)
    if curr_row is None:
        print("current row doesnt exist")
        return
    
    if bootstrap:
        curr_row, cluster_assignments_bootstrap = gmm_stats.bootstrap_gmm(curr_row,final_data[final_data.outlier == False], resample_size, CI_width, max_peaks, out)
    if allele_specif_CIs:
        for assignment in final_data['cluster_assignments'].unique():
            curr_row["H"+ str(assignment+1)+":median_CI_allele_specific"] = gmm_stats.bootstrap_gmm_allele_specific(final_data[final_data.cluster_assignments == assignment],resample_size,CI_width,out)

    #write out cluster assignments to file
    assignments = final_data[['read_id','counts']]
    assignments["Name"] = gmm_stats.name
    assignments['assignment'] = final_data['cluster_assignments']
    
    #check if this is the first entry, if it is, save the header, if its not, append to existing file
    if os.path.exists(out + "read_assignments_gmm.tsv"):
        assignments[['Name','read_id','counts','assignment']].to_csv(out + "read_assignments_gmm.tsv",header=None, sep="\t", index=False, mode="a")
    else:
        assignments[['Name','read_id','counts','assignment']].to_csv(out + "read_assignments_gmm.tsv", sep="\t", index=False, mode="a")

    return curr_row #return genotype


def main():
    parser = argparse.ArgumentParser()
    #Required inputs, TODO subcommands for input types, test this
    subparsers = parser.add_subparsers(dest='subcommand')
    subparsers.required = True
        #  subparser for dump
    parser_infile = subparsers.add_parser('targets_tsv')
    # add a required argument
    parser_infile.add_argument(dest ="targets",type=str, help='TSV with name, prefix, repeat, suffix, header required. Recommended at least 100bp with default alignment parameters')
    #coordinates mode parser
    parser_coords = subparsers.add_parser('coordinates')
    parser_coords.add_argument(dest = 'coords_file',type=str,help='Path to input bed file with either 4 (chr,start,end,repeat motif) or 5 (chr,start,end,repeat motif, name) columns')
    parser_coords.add_argument(dest='chrom_sizes_file',type=str,help = 'Path to chromosome sizes file')
    parser_coords.add_argument(dest="ref",type=str,help="Path to reference genome to get flanking sequence from")
    parser_coords.add_argument("--input_flank_length",type=int,help="Length of prefix and suffix to get from reference, must be longer than 30bp (Default) or flanking_size parameter (default: %(default)s)",default=200)
    #parser.add_argument(dest ="targets",type=str, help='TSV with name, prefix, repeat, suffix, header required. Recommended at least 100bp with default alignment parameters')
    parser.add_argument(dest ="out",type=str, help='Output prefix including directory path')
    parser.add_argument(dest = "inFile", type=str, help= 'Sequence to search and annotate in fasta or fastq, gzipped accepted')

    #coordinate input options ***these inputs are dependent on other inputs and I need to figure out how to deal with that
    #TODO figure out how to make arguments conditional ie how do I make it so you can pass either coordinates or target sheet? do I need a bunch of conditionals to control this or is there a
    #streamlined way to do this?

    #optional, if input, models and accompanying files already produced and use the provided prefix
    parser.add_argument("--background", type=str, help="Pickle of custom background dictionary, must include empty string for deletion character and total must add up to 1", default=None)
    parser.add_argument("--E_probs", type=str, help= 'TSV of custom emissions probability matrix, see format specifications', default=None)
    parser.add_argument("--A_probs", type=str, help= 'TSV of custom transition probabilities, see format specifications', default=None)
    parser.add_argument("--custom_RM", type=str, help= 'TSV of custom repeat match state probability matrix, see format specifications, used for motif mosacism, currently supports single target', default=None)
    parser.add_argument("--hmm_pre", type=str,help="Prefix for files produced by build function, use if running the same targets across multiple input files") #if I clean up the directory after I will get rid of this potentially
    #defaults subject to change
    parser.add_argument("--output_hist",help="Output supporting read histogram", action='store_true')
    #parser.add_argument("--exclude_sd", type=int,help="Number of standard deviations from the mean of tallest peaks to exclude neighboring peaks", default=2)
    parser.add_argument("--max_peaks", type=int,help="Maximum number of peaks to calculate AIC and BIC for in peak calling (default: %(default)s)",default=2)
    parser.add_argument("--cpus", type=int,help="Number of cpus to use. If none given, half of available cpus used (default: %(default)s)",default=int(mp.cpu_count()/2))
    parser.add_argument("--flanking_size", type=int,help="Number of basepairs flanking repeat to encode in model, longer recommended for highly repeatitive regions (default: %(default)s)",default=30)
    #alignment parameters
    parser.add_argument("--mode", type=str,help="map-ont (Nanopore), pb (PacBio), or sr (short accurate reads, use for accurate short flanking sequence input) (default: %(default)s)",default="map-ont")
    parser.add_argument("--cutoff", type=int,help="MapQ cutoff for prefix and suffix alignment (default: %(default)s)",default=30)
    parser.add_argument("--k", type=int,help="Kmer parameter to be passed to mappy",default=None)
    parser.add_argument("--w", type=int,help="Window parameter to be passed to mappy",default=None)
    parser.add_argument("--use_full_read", help="Flag to not subset the read by the alignment, optimal for reads with short prefix and suffix",action='store_true')

    #optional boostrapping parameters
    parser.add_argument("--bootstrap", help="Boolean designating to output boostrapped confidence intervals with genotype calls",action='store_true')
    parser.add_argument("--call_width", type=float, help="Decimal percentage designating width of confidence interval for allele calls (default: %(default)s)", default=0.95)
    parser.add_argument("--resample_size", type=int, help="Number of times to resample in bootstrapping (default: %(default)s)", default=500)

    #allele specific stats options
    parser.add_argument("--allele_specific_CIs",  help="Output allele-specific bootsrapped confidence intervals", action='store_true')
    parser.add_argument("--allele_specific_plots", help="Output allele-specific histograms with model of best fit", action='store_true')
    parser.add_argument("--discard_outliers", help="Discard outliers based on quantile", action='store_true')
    parser.add_argument("--filter_quantile", type=float, help="Float designating quantile of count frequency to discard when filtering outliers (default: %(default)s)", default=0.25)
    #for debugging
    parser.add_argument("--cluster_only", help="Run clustering on output corresponding to required arguments, mostly for debugging past runs", action='store_true')
    parser.add_argument("--save_intermediates", help="Flag designating to save intermediate files such as model inputs, raw count files, and state sequence files", action='store_true')

    args = parser.parse_args()

    #check input mode and check if inputs compatible for coordinates
    if args.subcommand == 'coordinates':
        #check inputs
        if args.input_flank_length < args.flanking_size:
            print("Prefix and suffix length (input_flank_length,%d) are incompatible with model flanking length (flanking_size, %d), exiting..."%(args.input_flank_length,args.flanking_size))
            return
        #generate infile from coordinates given
        print("Coordinates input selected, getting %dbp flanking sequence from reference at %s..."%(args.input_flank_length,args.ref))
        targets = generate_input_sheet(args.coords_file, args.chrom_sizes_file, args.ref, args.input_flank_length)
        #TODO decide if this should be an option
        print("Inputs generated from coordinates! Saving target inputs to %s..."%(args.out+"_inputs.tsv"))
        targets.to_csv(args.out + "_inputs.tsv", sep="\t",index=False)
    else:
        #read in targets
        targets = pd.read_csv(args.targets, sep="\t")
    #convert custom inputs
    if args.E_probs is not None:
        E_probs = read_model_params(args.E_probs,"emissions")
    else:
        E_probs = None
    if args.A_probs is not None:
        A_probs = read_model_params(args.A_probs,"transitions")
    else:
        A_probs = None
    if args.background is not None:
        background = read_model_params(args.background,"background")
    else:
        background = None
    if args.custom_RM is not None:
        custom_RM = read_model_params(args.custom_RM,"repeat")
        print(custom_RM)
    else:
        custom_RM = None

    #default alphabet that is compatible with all methods, subject to change if users want to have additional inputs such as methylated bases ect in the future
    alphabet = ['A','T','C','G','']
    if args.cluster_only == False:
        if args.hmm_pre is None:
            pool_start = perf_counter()
            targets.apply(build_all,axis=1,args=(args.out, background, alphabet,A_probs, E_probs, custom_RM, args.flanking_size)) #added flanking size
            pool_end = perf_counter()
            print("build_profile finished .... time was: ", str(pool_end-pool_start))
            build_pre = args.out #need to make sure this is compatible with new build all implentation
        else:
            print("Using previously written input files with prefix: " + args.hmm_pre)
            build_pre = args.hmm_pre
        convert_to_fasta(targets,args.out) #write out all target flanking sequences to fasta files to be used by mappy

        #file prefixes and suffixes to use
        hmm_file = ".hmm" #this is a suffix, does not include individual names****
        rev_hmm_file = "_revcomp" + ".hmm"
        hidden_states = ".hidden_states.txt"
        hidden_states_rev = "_revcomp.hidden_states.txt"

        #start multiprocess of read processing
        pool_start = perf_counter()
        print("cpu's in use: " + str(args.cpus))
        pool = mp.Pool(processes=args.cpus)

        #check input file type and pass to appropriate parser
        #check if compressed
        if args.inFile.endswith('gz'):
            #gzipped file
            if args.inFile.endswith("fasta.gz") or args.inFile.endswith("fa.gz"): #fasta file
                [pool.apply_async(process_read, args=(header,seq,hmm_file,rev_hmm_file,hidden_states,hidden_states_rev,args.out,targets, build_pre, args.mode, args.cutoff, args.k, args.w, args.use_full_read, args.flanking_size)) for header, seq, bool in read_fasta(gzip.open(args.inFile,'rt'))]
            elif args.inFile.endswith("fastq.gz") or args.inFile.endswith("fq.gz"): #fastq
                [pool.apply_async(process_read, args=(header,seq,hmm_file,rev_hmm_file,hidden_states,hidden_states_rev,args.out,targets, build_pre, args.mode, args.cutoff, args.k, args.w, args.use_full_read, args.flanking_size)) for header, seq, bool in read_fastq(gzip.open(args.inFile,'rt'))]
        else:
            if args.inFile.endswith('a'): #fasta file
                [pool.apply_async(process_read, args=(header,seq,hmm_file,rev_hmm_file,hidden_states,hidden_states_rev,args.out,targets, build_pre, args.mode, args.cutoff, args.k, args.w, args.use_full_read,args.flanking_size)) for header, seq, bool in read_fasta(open(args.inFile))]
            elif args.inFile.endswith('q'): #fastq
                [pool.apply_async(process_read, args=(header,seq,hmm_file,rev_hmm_file,hidden_states,hidden_states_rev,args.out,targets, build_pre, args.mode, args.cutoff, args.k, args.w, args.use_full_read,args.flanking_size)) for header, seq, bool in read_fastq(open(args.inFile))]
        pool.close()
        pool.join()
        pool_end = perf_counter()
        print("All reads processed, the pooled run took: ", pool_end-pool_start)

    #stats runs and allele calls
    if args.hmm_pre is None:
        build_pre = args.out #need to make sure this is compatible with new build all implentation
    else:
        build_pre = args.hmm_pre
    out_count_name = "_counts.txt"

    pool_start = perf_counter()
    #call original allele call procedure
    geno_df = targets.apply(ratio_gmm_stats,axis=1,args=(args.out,out_count_name, args.output_hist,args.max_peaks,args.discard_outliers, args.bootstrap, args.call_width, args.resample_size,args.allele_specific_plots,args.allele_specific_CIs))
    #geno_df.dropna(axis="rows",how="any", inplace=True)
    if isinstance(geno_df, pd.DataFrame):
        geno_df.to_csv(args.out + "genotype_calls_gmm.tsv",sep="\t",index=False)
        pool_end = perf_counter()
        print("GMM run done! Took: ", pool_end-pool_start)
    else:
        print(geno_df)
        print("Results are not a DataFrame! Something went wrong...")

    #clean up intermediates
    #TODO make this an option
    #all files produced for models
    if not args.save_intermediates:
        for file_subset in glob.glob(args.out+"*hidden_states.txt"):
            if os.path.exists(file_subset):
                os.remove(file_subset)
        for file_subset in glob.glob(args.out+"*.hmm"):
            if os.path.exists(file_subset):
                os.remove(file_subset)
        for file_subset in glob.glob(args.out+"*labeled_seqs.txt"):
            if os.path.exists(file_subset):
                os.remove(file_subset)
        #FIXME counts file is required for cluster only
        for file_subset in glob.glob(args.out+"*counts.txt"):
            os.remove(file_subset)
        if os.path.exists(args.out+"_prefix.fa"):
            os.remove(args.out+"_prefix.fa")
        if os.path.exists(args.out+"_suffix.fa"):
            os.remove(args.out+"_suffix.fa")
if __name__ == "__main__":
  main()


import seaborn as sns
from sklearn.mixture import GaussianMixture
from scipy.stats import norm
import scipy.stats as st
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import pickle as pkl
import math
from matplotlib.ticker import MaxNLocator
import warnings
warnings.filterwarnings('ignore')

from HMMSTR_utils.HMMSTR_utils import remove_outlier_IQR, throw_low_cov, remove_flanking_outliers

class GMMStats:
    def __init__(self, target_row):
        self.name = target_row[0]
        self.prefix = target_row[1].rstrip().upper()[-30:]
        self.repeat = target_row[2].rstrip().upper()
        self.suffix = target_row[3].rstrip().upper()[:30]
    
    def plot_stats(self,counts_data,out, strand=None):
        '''
        plot the read support histograms

        Args:
        -----------------------------------------------------------------------------
        counts_data (dataframe): dataframe continaing copy numbers from Viterbi output
        out (str): output prefix
        strand (str): strand we are plotting if applicable

        Returns:
        ----------------------------------------------------------------------------
        None

        '''
        to_plot = counts_data[['counts','freq']].drop_duplicates().sort_values(by="counts")
        if to_plot.shape[0] == 0 or to_plot.shape[1] == 0:
            return
        sns.reset_orig()
        to_plot.plot.bar(x="counts",y="freq",figsize=(15,5),title= str(self.name), color="red", fontsize=7)
        plt.xlabel("Repeat count")
        plt.ylabel("Number of reads")
        fig = plt.gcf()
        if strand is not None:
            fig.savefig(str(out + "_plots/"+self.name+ "_"+strand+"_supporting_reads_hist.pdf"))
        else:
            fig.savefig(str(out + "_plots/"+self.name+ "_supporting_reads_hist.pdf"))
        plt.show()
        plt.clf()
        return #currently does not work and isnt called
    def get_stats(self, out_count_file, out, plot_hists, filter_outliers=False, filter_quantile=0.25, flanking_like_filter=False, curr_strand=None):
        '''
        Function to calculate summary stats for a single target

        Args:
        -----------------------------------------------------------------------------
        out_count_name(str): output suffix for final result tsv
        out (str): output prefix
        plot_hists (bool): boolean indicating if histograms should be saved
        filter_outliers (bool): boolean indicating if outliers should be included or not
        filter_quantile (float): quantile that will be filtered if filter_outliers is true
        flanking_like_filter (bool): boolean designating if reads should be filtered based on the likelihood of their flanking sequence
        curr_strand (str): which strand we are performing operations on if applicable

        Returns:
        -----------------------------------------------------------------------------
        counts_data(pandas dataframe): final calculations for given count file

        '''
        #read output file
        counts_data = pd.read_csv(out_count_file, sep=" ",header=None)
        counts_data.columns = ["read_id","strand","align_score","neg_log_likelihood","subset_likelihood","repeat_likelihood","repeat_start","repeat_end","align_start", "align_end","counts"]
        
        if curr_strand is not None:
            counts_data = counts_data[counts_data.strand == curr_strand]
        
        counts_data['freq'] = counts_data.groupby(by='counts')['read_id'].transform('count')
        #outlier filtering
        outliers = []
        flanking_outliers = []
        if filter_outliers:
            filtered, outliers = remove_outlier_IQR(counts_data.counts, filter_quantile)
        counts_data['outlier'] = np.where(counts_data.counts.isin(outliers),True , False)
        if flanking_like_filter:
            flanking_filtered, flanking_outliers = remove_flanking_outliers(counts_data)
        counts_data['flanking_outlier'] = np.where(counts_data.read_id.isin(flanking_outliers),True , False)
        if plot_hists:
            self.plot_stats(counts_data, out, curr_strand)
        return counts_data

    def call_peaks(self,X_orig,out, num_max_peaks, plot=True, save_allele_plots = True, strand=None):
        '''
        Call peaks using a Gaussian mixture model. Code adapted from https://www.cbrinton.net/ECE20875-2020-Spring/W11/gmms_notebook.pdf

        Args:
        -----------------------------------------------------------------------------
        X_orig (dataframe): counts matrix
        out (str): output prefix
        num_max_peaks (int): maximum number of peaks to call
        plot (bool): boolean indicating if plots should be saved
        save_allele_plots (bool): boolean indicating if allele specific plots should be saved
        strand (str): which strand operations are performed on if applicable

        Returns:
        -----------------------------------------------------------------------------
        curr_row (Series): series containing all stats for current target
        cluster_assignments (Series): series containing read to cluster assignments
        '''
        rng = np.random.RandomState(seed=1)

        #get transformed counts column to represent weighted frequencies
        X = X_orig.copy()
        X = X[["counts"]]
        if X.shape[0] < 2:
            curr_dict = {"name":self.name}
            #check if we have 1 read
            if X.shape[0] == 1:
                try:
                    # curr_dict["H1:mean"] = X.counts[0]
                    curr_dict["A1:median"] = X.counts[0]
                    curr_dict["A1:mode"] = X.counts[0]
                    curr_dict["A1:SD"] = 0
                    curr_dict["A1:supporting_reads"] = 1
                    cluster_assignments = np.array([0])
                except:
                    print("X.counts[0] does not exist, continuing to write null row...")
                    cluster_assignments = np.array([-1])
            else:
                cluster_assignments = np.array([-1])
            for i in range(1,num_max_peaks+1):
                #curr_mean = "A"+str(i)+":mean"
                curr_median = "A"+str(i)+":median"
                curr_mode = "A"+str(i)+":mode"
                curr_sd = "A" + str(i) + ":SD"
                curr_support = "A" + str(i) + ":supporting_reads"
                if curr_median not in curr_dict.keys():
                    curr_dict[curr_median] = 0
                if curr_mode not in curr_dict.keys():
                    curr_dict[curr_mode] = 0
                if curr_sd not in curr_dict.keys():
                    curr_dict[curr_sd] = 0
                if curr_support not in curr_dict.keys():
                    curr_dict[curr_support] = 0
            curr_dict["num_supporting_reads"] = len(X_orig.index)
            return pd.Series(curr_dict), cluster_assignments
        #account for more components than samples
        if X.shape[0] < num_max_peaks:
            num_max_peaks = X.shape[0]

        k_arr = np.arange(num_max_peaks) + 1
        models = [
        GaussianMixture(n_components=k).fit(X)
        for k in k_arr
        ]
        dfs = [] #subsets of X corresponding to each peak

        # Plot function
        def plot_mixture(gmm, X, show_legend=True, ax=None, title=None, subplot=False,cluster=None):
            if plot:
                plt.rcParams["figure.figsize"] = (10,6)
                sns.set_style('whitegrid',{'grid.color': '#dcddde'}) 
                if ax is None:
                    ax = plt.gca()

            # Compute PDF of whole mixture
            x = np.linspace(X.min(), X.max(), 1000)
            logprob = gmm.score_samples(x.reshape(-1, 1))
            pdf = np.exp(logprob)

            # Compute PDF for each component
            responsibilities = gmm.predict_proba(x.reshape(-1, 1))

            pdf_individual = responsibilities * pdf[:, np.newaxis]

            #get the max of each pdf
            maxInColumns = np.amax(pdf_individual, axis=0)

            # Plot data histogram
            if plot:
                ax.xaxis.set_major_locator(MaxNLocator(integer=True))
                hist = ax.hist(X.counts, density=True, histtype='bar',  label='Data', bins=range(int(min(X.counts)),int(max(X.counts))+2),facecolor = '#9bc1cc', edgecolor='whitesmoke', linewidth=0.5, align="left") #updated plotting parameters
                ax.set_xlabel("Repeat Copy Number")
                ax.set_ylabel("Density")
                # Plot PDF of whole model
                if not subplot:
                    ax.plot(x, pdf, '-k', label='Mixture PDF',linewidth=2)

                # Plot PDF of each component, divided so we can call this multiple times, once for full mixture and once for each cluster
                if subplot:
                    ax.plot(x, pdf_individual[:,cluster],color="black")
                else:
                    ax.plot(x, pdf_individual, '--', label='Component PDF')
                if title is not None:
                    ax.set_title(title)
                if subplot:
                    #add metrics to plot
                    ymin, ymax = ax.get_ylim()
                    ax.set_ylim(ymin, ymax)
                    ax.vlines(x = st.mode(np.array(X.counts), keepdims=True)[0][0],ymin=ymin, ymax=ymax,linewidth=2, color='r',linestyle="dashed", label="Mode")
                    ax.axvline(x = X.counts.median(), linewidth=2, color='purple',linestyle="dashed",label="Median")
                if show_legend:
                    ax.legend()
                if not subplot: #save full peaks plot
                    if strand is not None:
                        plt.savefig(out + "_plots/"+ self.name +"_"+strand+ "peaks.pdf",bbox_inches='tight')
                    else:
                        plt.savefig(out + "_plots/"+ self.name + "peaks.pdf",bbox_inches='tight')
                    plt.show()
                    plt.clf()
                else:
                    if strand is not None:
                        plt.savefig(out + "_plots/"+ self.name + "_"+ strand+"allele_"+ str(i+1)+".pdf",bbox_inches='tight')
                    else:
                        plt.savefig(out + "_plots/"+ self.name + "allele_"+ str(i+1)+".pdf",bbox_inches='tight')
                    plt.show()
                    plt.clf()
                plt.show()
                plt.clf()
            return gmm.means_[np.argsort(maxInColumns)],np.argsort(maxInColumns)
        AIC = [m.aic(X) for m in models]
        BIC = [m.bic(X) for m in models]

        # Plot these metrics
        if plot:
            fig = plt.gcf()
            fig.set_size_inches(15,7)
            plt.plot(k_arr, AIC, label='AIC')
            plt.plot(k_arr, BIC, label='BIC')
            plt.xlabel('Number of Components ($k$)')
            plt.legend()
            plt.tight_layout()
            plt.title(self.name+": AIC and BIC")
            if strand is not None:
                plt.savefig(out + "_plots/"+ self.name +"_"+strand+ "_AIC_BIC.pdf",bbox_inches='tight')
            else:
                plt.savefig(out + "_plots/"+ self.name +"_AIC_BIC.pdf",bbox_inches='tight')
            plt.show()
            plt.clf()
            fig = plt.gcf()
            fig.set_size_inches(15,7)
        gmm_best = models[np.argmin(AIC)]
        max_peaks, ranks_idx = plot_mixture(gmm_best, X, title=self.name +": Gaussian Mixture Model of Best Fit")

        cluster_assignments = np.argmax(gmm_best.predict_proba(X), axis=1)
        for cluster in np.unique(cluster_assignments):
            dfs.append(X[np.where(cluster_assignments == cluster, True, False)])
        #plot separate clusters
        curr_dict = {"name":self.name}
        cluster_stats = {}
        for i,cluster in enumerate(dfs):
            #plot distributions separately
            cluster_stats[i] = {}
            if save_allele_plots and plot:
                test1, test2 = plot_mixture(gmm_best, cluster,title=self.name+": Model of Best Fit for Allele " + str(i+1), subplot=True,cluster=i)
            #add metrics to dictionary so they are outputted
            cluster_stats[i]['median'] = cluster.counts.median()
            cluster_stats[i]['mode'] = st.mode(np.array(cluster),keepdims=True)[0][0][0]
            if math.isnan(cluster.counts.std()):
                cluster_stats[i]['SD']  = 0
            else:
                cluster_stats[i]['SD'] = cluster.counts.std()
            cluster_stats[i]['supporting_reads'] = len(np.where(cluster_assignments == i)[0])
        #unravel cluster_stats dictionary for curr_row
        for i in cluster_stats.keys():
            #curr_dict["A" + str(i+1) + ":mean"] = cluster_stats[i]["mean"]
            curr_dict["A" + str(i+1) + ":median"] = cluster_stats[i]["median"]
            curr_dict["A" + str(i+1) + ":mode"] = cluster_stats[i]["mode"]
            curr_dict["A" + str(i+1) + ":SD"] = cluster_stats[i]["SD"]
            curr_dict["A" + str(i+1) + ":supporting_reads"] = cluster_stats[i]["supporting_reads"]

        #check for missing columns and fill them if not there
        for i in range(1,num_max_peaks+1):
            #curr_mean = "A"+str(i)+":mean"
            curr_median = "A"+str(i)+":median"
            curr_mode = "A"+str(i)+":mode"
            curr_sd = "A" + str(i) + ":SD"
            curr_support = "A" + str(i) + ":supporting_reads"
            #if curr_mean not in curr_dict.keys():
                #curr_dict[curr_mean] = 0
            if curr_median not in curr_dict.keys():
                curr_dict[curr_median] = 0
            if curr_mode not in curr_dict.keys():
                curr_dict[curr_mode] = 0
            if curr_sd not in curr_dict.keys():
                curr_dict[curr_sd] = 0
            if curr_support not in curr_dict.keys():
                curr_dict[curr_support] = 0
        curr_dict["num_supporting_reads"] = len(X_orig.index)
            
        curr_row = pd.Series(curr_dict)
        return curr_row, cluster_assignments
    
    def bootstrap_gmm(self,curr_row,data, resample_size, CI_width, max_peaks, out):
        '''
        Performs bootstrapping and constructs confidence intervals using GMM peakcalling

        Args:
        --------------------------------------------------------------------------------
        curr_row (series): current row in the genotype dataframe to append CIs to
        data (dataframe): counts dataframe to sample from
        resample_size (int): number of times to resample during bootstraping
        CI_width (float): width of confidience interval to calculate
        max_peaks (int): maximum number of peaks to call in each bootstrap iteration
        out (str): output prefix

        Returns:
        --------------------------------------------------------------------------------
        curr_row (series): the original input row with bootstrap results included
        '''
        #throw out 1 read coverage to avoid skewed bootstrapping results
        filtered = throw_low_cov(data)
        call_lists = {}
        for i in range(1,max_peaks+1):
            call_lists["A"+str(i)+"_median"]=[]

        for i in range(resample_size):
            if resample_size > 100 and (i+1)%10 == 0:
                print("Round " + str(i+1)+ "/" + str(resample_size))
            elif resample_size < 100 and resample_size > 10 and (i+1)%5 == 0:
                print("Round " + str(i+1)+ "/" + str(resample_size))
            elif resample_size < 10:
                print("Round " + str(i+1)+ "/" + str(resample_size))
            #sample up to the depth of the experiment with replacement
            curr_sample = filtered.sample(frac=1,axis=0, replace = True)
            #call gmm with original freqs on this subset
            curr_row = self.call_peaks(curr_sample, out ,max_peaks, plot=False) 
            #add current call to lists
            for j in range(1,max_peaks+1):
                call_lists["A"+str(j)+"_median"].append(curr_row[0]['A'+str(j)+":median"])
        adjustment = float((1.0 - CI_width)/2)
        for j in range(1,max_peaks+1): #get confidence interval for each allele called
            curr_medians = call_lists["A"+str(j)+"_median"]
            curr_row[0]["A"+str(j)+":median_CI"] = (np.quantile(curr_medians,adjustment),np.quantile(curr_medians,1-adjustment))
        return curr_row[0]

    def bootstrap_gmm_allele_specific(self,data, resample_size, CI_width, out):
            '''
            Performs bootstrapping and constructs confidence interval for single allele using GMM peakcalling

            Args:
            --------------------------------------------------------------------------------
            data (dataframe): counts dataframe to sample from
            resample_size (int): number of times to resample during bootstraping
            CI_width (float): width of confidience interval to calculate
            out (str): output prefix

            Returns:
            --------------------------------------------------------------------------------
            median_CI (tuple): tuple containing the median confidence interval call for the current allele 
            '''
            #throw out 1 read coverage, this is something to assume will happen if bootstrapping
            filtered = throw_low_cov(data)
            call_lists = {}
            #call_lists["mean"]=[]
            call_lists["median"]=[]
            #call_lists["mode"]=[]

            for i in range(resample_size):
                if resample_size > 100 and (i+1)%10 == 0:
                    print("Round " + str(i+1)+ "/" + str(resample_size))
                elif resample_size < 100 and resample_size > 10 and (i+1)%5 == 0:
                    print("Round " + str(i+1)+ "/" + str(resample_size))
                elif resample_size < 10:
                    print("Round " + str(i+1)+ "/" + str(resample_size))
                #sample up to the depth of the experiment with replacement
                curr_sample = filtered.sample(frac=1,axis=0, replace = True)
                #call gmm with original freqs on this subset
                curr_row = self.call_peaks(curr_sample, out ,1, plot=False) #single allele bootstrapping
                #add current call to lists
                call_lists["median"].append(curr_row[0]["A1:median"])
            adjustment = float((1.0 - CI_width)/2)
            curr_medians = call_lists["median"]
            median_CI= (np.quantile(curr_medians,adjustment),np.quantile(curr_medians,1-adjustment))
            return median_CI


import hmm as hmm
import argparse

#read in inputs for run_main
parser = argparse.ArgumentParser()
parser.add_argument("model_file", type=str)
parser.add_argument("seq_len", type=str)
parser.add_argument("seq", type=str)
parser.add_argument("repeat_len", type=str)
parser.add_argument("prefix_idx", type=str)
args = parser.parse_args()
hmm.run_main(args.model_file,args.seq_len,args.seq,args.repeat_len,args.prefix_idx)




from libc.stdio cimport FILE
cdef extern from "glib.h":
    ctypedef struct _GList:
        void* data
        _GList* next
        _GList* prev

    ctypedef _GList* GList

cdef extern from "glist.h":
    # Doubly linked lists
    GList* g_list_alloc() 
    void g_list_free(GList* list)
    void g_list_free_1(GList* list)
    void g_list_free_full(GList* list, void* free_func)
    GList* g_list_append(GList* list, void* data)
    GList* g_list_prepend(GList* list, void* data)
    GList* g_list_insert(GList* list, void* data, int position)
    GList* g_list_insert_sorted(GList* list, void* data, void* func)
    GList* g_list_insert_sorted_with_data(GList* list, void* data, void* func, void* user_data)
    GList* g_list_insert_before(GList* list, GList* sibling, void* data)
    GList* g_list_insert_before_link(GList* list, GList* sibling, GList* link_)
    GList* g_list_concat(GList* list1, GList* list2)
    GList* g_list_remove(GList* list, void* data)
    GList* g_list_remove_all(GList* list, void* data)
    GList* g_list_remove_link(GList* list, GList* llink)
    GList* g_list_delete_link(GList* list, GList* link_)
    GList* g_list_reverse(GList* list)
    GList* g_list_copy(GList* list)
    GList* g_list_copy_deep(GList* list, void* func, void* user_data)
    GList* g_list_nth(GList* list, unsigned int n)
    GList* g_list_nth_prev(GList* list, unsigned int n)
    GList* g_list_find(GList* list, void* data)
    GList* g_list_find_custom(GList* list, void* data, void* func)
    int g_list_position(GList* list, GList* llink)
    int g_list_index(GList* list, void* data)
    GList* g_list_last(GList* list)
    GList* g_list_first(GList* list)
    unsigned int g_list_length(GList* list)
    void g_list_foreach(GList* list, void* func, void* user_data)
    GList* g_list_sort(GList* list, void* compare_func)
    GList* g_list_sort_with_data(GList* list, void* compare_func, void* user_data)
    void* g_list_nth_data(GList* list, unsigned int n)
    void g_clear_list(GList** list_ptr, void* destroy)

#define g_list_previous(list) ((list) ? (((GList *)(list))->prev) : NULL)
#define g_list_next(list) ((list) ? (((GList *)(list))->next) : NULL)
cdef extern from "nrutil.h":
    float *vector(int nl,int nh)
    float **matrix(int nrl,int nrh,int ncl,int nch)
    float **convert_matrix()
    double *dvector(int nl,int nh)
    double **dmatrix(int nrl,int nrh,int ncl,int nch)
    int *ivector(int nl, int nh)
    int **imatrix(int nrl,int nrh,int ncl,int nch)
    float **submatrix(float** a,int oldrl,int oldrh,int oldcl,int oldch,int newrl,int newcl)
    void free_vector(float* v,int nl, int nh)
    void free_dvector(double* v, int nl, int nh)
    void free_ivector(int* v,int nl,int nh)
    void free_matrix(float** m,int nrl, int nrh, int ncl, int nch)
    void free_dmatrix(double** m, int nrl, int nrh, int ncl, int nch)
    void free_imatrix(int** m, int nrl, int nrh, int ncl, int nch)
    void free_submatrix(float** b, int nrl, int nrh, int ncl, int nch)
    void free_convert_matrix(float* a, int nrl, int nrh, int ncl, int nch)
    void nrerror(char errortext[])
cdef extern from "hmm.h":
    ctypedef struct HMM:
        int N
        int M
        double** A
        double** B
        double* pi

    void ReadHMM(FILE* fp, HMM* phmm)
    void PrintHMM(FILE* fp, HMM* phmm)
    void InitHMM(HMM* phmm, int N, int M, int seed)
    void CopyHMM(HMM* phmm1, HMM* phmm2)
    void FreeHMM(HMM* phmm)
    void PrintDelta(FILE* fp, double** delta, HMM* phmm, int T)
    void PrintTraceBack(FILE* fp, int** delta, HMM* phmm, int T)
    void ReadSequence(FILE* fp, int* pT, int** pO)
    void ReadSequences(FILE* fp, int* pN, int** pT, int*** pO)
    void PrintSequences(int* pN, int* size, int*** O)
    void PrintSequence(FILE* fp, int T, int* O)
    void PrintPath(FILE* fp, int T_new, GList* list)
    void PrintLikelihoods(FILE* fp, int T_new, GList* likelihoods)
    void GenSequenceArray(HMM* phmm, int seed, int T, int* O, int* q)
    int GenInitalState(HMM* phmm)
    int GenNextState(HMM* phmm, int q_t)
    int GenSymbol(HMM* phmm, int q_t)
    void PrintAllOuts(FILE* fp, int T_new, GList* list, double** delta, int N, int T)

    void Viterbi(HMM* phmm, int T, int* O, double** delta, int** psi, int** traceback_dir, GList** list, int* T_new, int* q, double* pprob)
    void ViterbiLog(HMM* phmm, int T, int* O, double** delta, int** psi, int** traceback_dir, GList** list, GList** likelihoods_ints, GList** likelihoods_dec, int* T_new, int* q, double* pprob, int repeat_len, int prefix_idx)

    int hmmgetseed()
    void hmmsetseed(int seed)
    double hmmgetrand()

    # Macros
    # Note: These will be inlined by the C compiler
    int MAX(int x, int y)
    int MIN(int x, int y)

# Wrapper functions for the HMM functions
cdef void py_ReadHMM(FILE* fp, HMM* phmm)
cdef void py_PrintHMM(FILE* fp, HMM* phmm)
cdef void py_InitHMM(HMM* phmm, int N, int M, int seed)
cdef void py_CopyHMM(HMM* phmm1, HMM* phmm2)
cdef void py_FreeHMM(HMM* phmm)
cdef void py_PrintDelta(FILE* fp, double** delta, HMM* phmm, int T)
cdef void py_PrintTraceBack(FILE* fp, int** delta, HMM* phmm, int T)
cdef void py_ReadSequence(FILE* fp, int* pT, int** pO)
cdef void py_ReadSequences(FILE* fp, int* pN, int** pT, int*** pO)
cdef void py_PrintSequences(int* pN, int* size, int*** O)

cdef void py_PrintSequence(FILE* fp, int T, int* O)
cdef void py_PrintPath(FILE* fp, int T_new, GList* list)
cdef void py_PrintLikelihoods(FILE* fp, int T_new, GList* likelihoods)
cdef void py_GenSequenceArray(HMM* phmm, int seed, int T, int* O, int* q)
cdef int py_GenInitalState(HMM* phmm)
cdef int py_GenNextState(HMM* phmm, int q_t)
cdef int py_GenSymbol(HMM* phmm, int q_t)
cdef void py_PrintAllOuts(FILE* fp, int T_new, GList* list, double** delta, int N, int T)

cdef void py_Viterbi(HMM* phmm, int T, int* O, double** delta, int** psi, int** traceback_dir, GList** list, int* T_new, int* q, double* pprob)

cdef void py_ViterbiLog(HMM* phmm, int T, int* O, double** delta, int** psi, int** traceback_dir, GList** list, GList** likelihoods_ints, GList** likelihoods_dec, int* T_new, int* q, double* pprob, int repeat_len, int prefix_idx)

cdef int py_hmmgetseed()
cdef void py_hmmsetseed(int seed)
cdef double py_hmmgetrand()
# Wrapper for the `free` function
cdef extern from "stdlib.h":
    void free(void* ptr)

cdef void py_free(void* ptr)



import pickle as pkl
from collections import defaultdict
#custom imports
from HMMSTR_utils.HMMSTR_utils import rev_comp
from importlib_resources import files

class ProfileHMM:
    def __init__(self, name,alphabet,prefix,repeat,suffix,background=None,out=".",transitions=None, mismatch_probs=None, repeat_probs =None,flanking_size=30):
        self.out = out
        self.name = name
        self.alphabet = alphabet
        self.flanking_size = flanking_size
        #initialize forward target characteristics
        self.prefix = prefix.upper()[-flanking_size:]
        self.repeat = repeat.upper()
        self.suffix = suffix.upper()[:flanking_size]
        #initialize reverse target characteristics
        self.rev_prefix = rev_comp(prefix.upper())[:flanking_size] #opposite subsetting for revcomp
        self.rev_repeat = rev_comp(repeat.upper())
        self.rev_suffix = rev_comp(suffix.upper())[-flanking_size:]
        self.get_states()
        
        #If no background inputted, default to uniform distribution
        est_072022_transitions_path = files('profile_HMM.data').joinpath('est_072022_transitions.pkl')
        est_072022_emissions_path = files('profile_HMM.data').joinpath('est_072022_emissions.pkl')
        if background is None:
            self.background = {'A':0.25,'T':0.25,'C':0.25,'G':0.25,'':0}
        else:
            self.background = background
        if transitions is None:
            #read default pkl file
            with open(est_072022_transitions_path,"rb") as file: #assumes defaults are in same directory as scripts, may need to change this
                self.transitions = pkl.load(file)
        else:
            self.transitions = transitions
        if mismatch_probs is None:
            with open(est_072022_emissions_path, "rb") as file:
                self.emissions = pkl.load(file)
        else:
            self.emissions = mismatch_probs
        if repeat_probs is None:
            self.repeat_probs = None
        else:
            self.repeat_probs = repeat_probs

    def get_states(self):
        '''
        Function to calculate the number of match states needed
        Returns:
            states (int): number of match states

        '''
        states = 0
        states += len(self.prefix) + len(self.repeat) + len(self.suffix)
        self.states = states

    def build_profile(self, strand):
        '''
        Function to initialize a Profile HMM structure

        Args:
            strand: which strand to build model with reference to
        Returns:
            None, writes HMM file in format needed by umdhmm-v1.02

        '''
        #set prefix, suffix, repeat according to strand
        if strand == "forward":
            prefix = self.prefix
            repeat = self.repeat
            suffix = self.suffix
            if self.repeat_probs is not None:
                repeat_probs = self.repeat_probs
            else:
                repeat_probs = None
        else:
            prefix = self.rev_suffix
            repeat = self.rev_repeat
            suffix = self.rev_prefix
            if self.repeat_probs is not None:
                    conversion = {'A':'T','G':'C','T':'A','C':'G','':''}
                    conversion_states = dict(zip(list(self.repeat_probs.keys()),list(self.repeat_probs.keys())[::-1]))
                    repeat_probs = {}
                    for key in self.repeat_probs.keys():
                        repeat_probs[key] = {}
                        for bp in self.repeat_probs[key]:
                            repeat_probs[key][bp] = self.repeat_probs[conversion_states[key]][conversion[bp]]
            else:
                repeat_probs = None
        #Initialize empty matrices A, E, and I
        A = defaultdict(dict)
        E = defaultdict(dict)
        I = {}
        hidden_states = []

        #initialize 'G' state at start to absorb bases before prefix starts
        hidden_states.append('G1')

        #initialize prefix states
        for i in range(1,len(prefix)+1):
            hidden_states.append('PD'+str(i))
            hidden_states.append('PI'+str(i))
            hidden_states.append('PM'+str(i))

        #initialize repeat states
        for i in range(1,len(repeat)+1):
            hidden_states.append('RD'+str(i))
            hidden_states.append('RI'+str(i))
            hidden_states.append('RM'+str(i))

        #initialize suffix states
        for i in range(1,len(suffix)+1):
            #we don't want an insert in the last position
            hidden_states.append('SD'+str(i))
            if i < len(suffix):
                hidden_states.append('SI'+str(i))
            hidden_states.append('SM'+str(i))
        #initialize 2nd 'G' state
        hidden_states.append('G2')

        hidden_states.append('E') #end node

        # Initialize all transitions to 0
        for item in hidden_states:
            for next_item in hidden_states:
                A[item][next_item] = 0.0

        #initialize start state
        for item in hidden_states:
            I[item] = 0.0

        #need to test probabilities here
        I['G1'] = 1.0

        #initialize emission probabilities
        E = self.get_emission(E, prefix, repeat, suffix,repeat_probs)
        A = self.get_transitions(A, prefix, repeat, suffix)
        if strand == "reverse":
            out = self.out + "_" + self.name + "_revcomp"
            self.rev_hidden_states = hidden_states
            self.rev_A = A
            self.rev_E = E
            self.rev_I = I
        else:
            out = self.out + "_" + self.name
            self.hidden_states = hidden_states
            self.A = A
            self.E = E
            self.I = I
        #call with current A,E,I since then we don't have to change the call based on strand but we have them saved correctly
        self.write_HMM(hidden_states, A, E, I, out)

        return   
    def get_emission(self,E, prefix, repeat, suffix,repeat_probs):
        '''
        Function to initialize emission probability matrix

        Args:
            E (defaultdict): empty default dictionary to fill
            prefix (str): prefix from reference genome to match to
            repeat (str): repeat motif to match
            suffix (str): suffix from reference to match to
            repeat_probs (dictionary): custom repeat probability matrix from user input (if applicable)
        Returns:
            E (defaultdict): filled emissions matrix

        '''
        #initialize 'G' states to have background nucleotide frequencies
        E['G1'] = self.background
        E['G2'] = self.background

        #i don't know if I need this but I'm filling this out as silent just in case
        for bp in self.alphabet:
            E['E'][bp] = 0.0
        #initialize prefix emissions based on prefix sequence
        for i in range(1, len(prefix)+1):
            for bp in self.alphabet:
                E['PD' + str(i)][bp] = 0.0 #deletion states don't emit a base
                E['PI' + str(i)][bp] = self.background[bp]
                if prefix[i-1] == bp:
                    E['PM' + str(i)][bp] = self.emissions[bp][bp]#probability of match
                elif bp == "":
                    E['PM' + str(i)][bp] = 0.0
                else:
                    E['PM' + str(i)][bp] = self.emissions[prefix[i-1]][bp] #probability of mismatch error
            E['PD' + str(i)][""] = 0.25 #try with background freqs

        #initialize repeat emissions based on repeat sequence
        #if custom probabilities given, add dictionary to full emissions dictionary, else continue as usual with emission probabilities
        if repeat_probs is None:
            for i in range(1,len(repeat)+1):
                for j,bp in enumerate(self.alphabet):
                    E['RD' + str(i)][bp] = 0.0 #deletion states don't emit a base
                    E['RI' + str(i)][bp] = self.background[bp] #might need to update this when I update background frequencies

                    if repeat[i-1] == 'N':
                        if bp in ['A','C','T','G']:
                            E['RM' + str(i)][bp] = 0.25 #set ptobaility to 0.25 for all bases since it can be any of them
                        else:
                            E['RM' + str(i)][bp] = 0.0
                    elif repeat[i-1] == bp:
                        E['RM' + str(i)][bp] = self.emissions[bp][bp] #probability of match
                    elif bp == "":
                        E['RM' + str(i)][bp] = 0.0
                    else:
                        E['RM' + str(i)][bp] = self.emissions[repeat[i-1]][bp] #probability of mismatch error
                E['RD' + str(i)][""] = 0.25 #background freqs
        else: #load dictionary and proceed with all other state types as usual
            E.update(repeat_probs) #add ALL match state probs for the repeat region
            for i in range(1,len(repeat)+1):
                for j,bp in enumerate(self.alphabet):
                    E['RD' + str(i)][bp] = 0.0 #deletion states don't emit a base
                    E['RI' + str(i)][bp] = self.background[bp] #might need to update this when I update background frequencies
                E['RD' + str(i)][""] = 0.25 #try with background freqs
        #initialize the suffix emissions based on the suffix sequence
        for i in range(1,len(suffix)+1):
            for j,bp in enumerate(self.alphabet):
                E['SD' + str(i)][bp] = 0.0 #deletion states don't emit a base
                if i < len(suffix): #skip last insertion states since its replaced by 'G2'
                    E['SI' + str(i)][bp] = self.background[bp] #might need to update this when I update background frequencies
                if suffix[i-1] == bp:
                    E['SM' + str(i)][bp] = self.emissions[bp][bp] #probability of match
                elif bp == "":
                    E['SM' + str(i)][bp] = 0.0
                else:
                    E['SM' + str(i)][bp] = self.emissions[suffix[i-1]][bp] #probability of mismatch error
            E['SD' + str(i)][""] = 0.25 #try with background freqs
        
        return E
    def get_transitions(self,A, prefix, repeat, suffix):
        '''
        Function to fill out all non-zero transitions
        Args:
            A (defaultdict): initialized transition matrix
            prefix (str): prefix from reference genome to match to
            repeat (str): repeat motif to match
            suffix (str): suffix from reference to match to


        Returns:
            A (defaultdict): filled out transition matrix

        '''
        #initialize 'G' transitions ***THIS IS A GUESS, MAY NEED TO STILL DECREASE
        A['G1']['G1'] = 0.50 
        A['G1']['PD1'] = 0.15
        A['G1']['PM1'] = 0.35

        #really unsure about these probabilities ***MAY NEED TO STILL CHANGE
        A['G2']['G2'] = 0.50
        A['G2']['E'] = 0.50

        #prefix transitions
        for i in range(1, len(prefix)+1): #all but the last Transitions
            #transitions into insertion states
            A['PI' + str(i)]['PI' + str(i)] = self.transitions["P_ii"]
            A['PD' + str(i)]['PI' + str(i)] = self.transitions["P_di"]
            A['PM' + str(i)]['PI' + str(i)] = self.transitions["P_mi"]
            #transition into deletion states
            if i > 1:
                A['PD'+str(i-1)]['PD'+str(i)] = self.transitions["P_dd"]  # from previous deletion
                A['PM'+str(i-1)]['PD'+str(i)] = self.transitions["P_md"]
                A['PI'+str(i-1)]['PD'+str(i)] = self.transitions["P_id"]
            #transition into match states
            if i > 1:
                A['PD'+str(i-1)]['PM'+str(i)] = self.transitions["P_dm"] # from previous deletion
                A['PM'+str(i-1)]['PM'+str(i)] = self.transitions["P_mm"]
                A['PI'+str(i-1)]['PM'+str(i)] = self.transitions["P_im"]
        #prefix border transitions
        A['PD' + str(len(prefix))]['RM1'] = self.transitions["P_dm"]
        A['PD' + str(len(prefix))]['RD1'] = self.transitions["P_dd"]
        A['PD' + str(len(prefix))]['PI' + str(len(prefix))] = self.transitions["P_di"]

        A['PI' + str(len(prefix))]['PI' + str(len(prefix))] = self.transitions["P_ii"]
        A['PI' + str(len(prefix))]['RM1'] = self.transitions["P_im"]
        A['PI' + str(len(prefix))]['RD1'] = self.transitions["P_id"]

        A['PM' + str(len(prefix))]['RM1'] = self.transitions["P_mm"]
        A['PM' + str(len(prefix))]['RD1'] = self.transitions["P_md"]
        A['PM' + str(len(prefix))]['PI' + str(len(prefix))] = self.transitions["P_mi"]

        #repeat transitions
        for i in range(1, len(repeat)+1): #all but the last Transitions
            #transitions into insertion states
            A['RI' + str(i)]['RI' + str(i)] = self.transitions["P_ii"]
            A['RD' + str(i)]['RI' + str(i)] = self.transitions["P_di"]
            A['RM' + str(i)]['RI' + str(i)] = self.transitions["P_mi"]
            #transition into deletion states
            if i > 1:
                A['RD'+str(i-1)]['RD'+str(i)] = self.transitions["P_dd"]  # from previous deletion
                A['RM'+str(i-1)]['RD'+str(i)] = self.transitions["P_md"]
                A['RI'+str(i-1)]['RD'+str(i)] = self.transitions["P_id"]
            #transition into match states
            if i > 1:
                A['RD'+str(i-1)]['RM'+str(i)] = self.transitions["P_dm"] # from previous deletion
                A['RM'+str(i-1)]['RM'+str(i)] = self.transitions["P_mm"]
                A['RI'+str(i-1)]['RM'+str(i)] = self.transitions["P_im"]
        #border and looping transitions (I haven't looked at this yet and it probably follows some distribution but thats unclear rn)
        A['RM' + str(len(repeat))]['SM1'] = 0.42
        A['RM' + str(len(repeat))]['SD1'] = 0.01
        A['RM' + str(len(repeat))]['RI' + str(len(repeat))] = 0.015
        #loop back from end of repeat
        A['RM' + str(len(repeat))]['RM1'] = 0.54
        A['RM' + str(len(repeat))]['RD1'] = 0.015
        #from deletion state
        A['RD' + str(len(repeat))]['SD1'] = 0.002
        A['RD' + str(len(repeat))]['SM1'] = 0.43
        A['RD' + str(len(repeat))]['RI' + str(len(repeat))] = 0.004
        #loop back from deletion
        A['RD' + str(len(repeat))]['RD1'] = 0.004
        A['RD' + str(len(repeat))]['RM1'] = 0.56
        #from insertion
        A['RI' + str(len(repeat))]['RI' + str(len(repeat))] = 0.004
        A['RI' + str(len(repeat))]['SM1'] = 0.43
        A['RI' + str(len(repeat))]['SD1'] = 0.002
        #loop back from insertion
        A['RI' + str(len(repeat))]['RM1'] = 0.56
        A['RI' + str(len(repeat))]['RD1'] = 0.004

        #suffix transitions
        for i in range(1, len(suffix)+1): #all but the last Transitions
            #transitions into insertion states
            if i != len(suffix): # there is no last insertion state
                A['SI' + str(i)]['SI' + str(i)] = self.transitions["P_ii"]
                A['SD' + str(i)]['SI' + str(i)] = self.transitions["P_di"]
                A['SM' + str(i)]['SI' + str(i)] = self.transitions["P_mi"]
            #transition into deletion states
            if i > 1:
                A['SD'+str(i-1)]['SD'+str(i)] = self.transitions["P_dd"]# from previous deletion
                A['SM'+str(i-1)]['SD'+str(i)] = self.transitions["P_md"]
                A['SI'+str(i-1)]['SD'+str(i)] = self.transitions["P_id"]
            #transition into match states
            if i > 1:
                A['SD'+str(i-1)]['SM'+str(i)] = self.transitions["P_dm"] # from previous deletion
                A['SM'+str(i-1)]['SM'+str(i)] = self.transitions["P_mm"]
                A['SI'+str(i-1)]['SM'+str(i)] = self.transitions["P_im"]
        #border suffix transitions
        A['SM' + str(len(suffix))]['G2'] = 1.0
        A['SD' + str(len(suffix))]['G2'] = 1.0
        return A   
    def write_HMM(self, hidden_states, A, B, pi, out):
        '''
        Function to write out Profile HMM in format needed by umdhmm
        Args:
            alphabet (list): alphabet characters for the model
            A (defaultdict(defaultdict)): transition matrix of HMM
            B (defaultdict(defaultdict)): emission probability matrix of HMM
            pi (defaultdict): initial probability matrix of HMM
            out (str): output file prefix to save to

        Returns:
            None, saves to out.hmm and out.hidden_states.txt
        '''
        M = len(self.alphabet)#length of alphabet
        N = self.states*3 + 2 #number of states including 'G', 'E', insertions and deletions
        with open(out + ".hidden_states.txt",'w') as outfile:
            outfile.write(".".join(hidden_states))
        with open(out + ".hmm", 'w') as outfile:
            outfile.write("M= " + str(M) + "\n")
            outfile.write("N= " + str(N) + "\n")
            outfile.write("A:\n")
            #loop across A to get transition matrix
            #use hidden states order since states were added in order of graph
            for curr_key in hidden_states:
                for curr_item in hidden_states: #loop across all keys in the inner dictionary
                    outfile.write(str(A[curr_key][curr_item])+ " ")
                outfile.write("\n")
            #write out the emission probability matrix
            outfile.write("B:\n")
            for curr_key in hidden_states:
                for bp in self.alphabet:
                    outfile.write(str(B[curr_key][bp])+ " ")
                outfile.write("\n")
            #write out pi
            outfile.write("pi:\n")
            for curr_key in hidden_states:
                outfile.write(str(pi[curr_key]) + " ")
            outfile.write("\n")
        return
    
        


import pandas as pd
import mappy
import numpy as np
def rev_comp(strand): #can refactor this to use dictionaries
    comp = ""
    for i in range(0, len(strand)):
        next_base = strand[len(strand) - i - 1]
        #check what the base is to switch to complement
        if next_base == "A":
            comp = comp + "T"
        if next_base == "T":
            comp = comp + "A"
        if next_base == "G":
            comp = comp + "C"
        if next_base == "C":
            comp = comp + "G"
        if next_base == "N":
            comp = comp + "N"
    return comp

def seq2int(seq):
  d = {'A': '1', 'T': '2', 'C': '3', 'G': '4'}
  seqInt = ' '.join(d[s] if s in d else s for s in seq.upper())
  #print(seqInt)
  return seqInt
  
def read_fasta(fileObject):
  '''
  Generator function to read in reads
  '''
  header = ''
  seq = ''
  # skip any useless leading information
  for line in fileObject:
    if line.startswith('>'):
      header = line.strip()
      break
  for line in fileObject:
    if line.startswith('>'):
      if 'N' in seq.upper():
        yield header, seq, False
      else:
        yield header, seq, True
      header = line.strip()
      seq = ''
    else:
      seq += line.strip()
  if header:
    if 'N' in seq.upper():
      yield header, seq, False
    else:
      yield header, seq, True

def read_fastq(fileObject):
  '''
  Generator function to read in reads
  '''
  header = ''
  seq = ''
  # skip any useless leading information
  for line in fileObject:
    print(line)
    if line.startswith('@'):
      header = line.strip()
      break
  for line in fileObject:
    if line.startswith('@'):
      if 'N' in seq.upper():
        yield header, seq, False
      else:
        yield header, seq, True
      header = line.strip()
      seq = ''
    #check if optional or quality field
    elif line.startswith('+'):
       #assume quailty is one line and is after optional line
       next(fileObject)
       continue
    else:
      seq += line.strip()
  if header:
    if 'N' in seq.upper():
      yield header, seq, False
    else:
      yield header, seq, True

def remove_outlier_IQR(df):
    Q1=df.quantile(0.25)
    Q3=df.quantile(0.75)
    IQR=Q3-Q1
    df_final=df[~((df<(Q1-1.5*IQR)) | (df>(Q3+1.5*IQR)))]
    outliers= df[~(df.isin(df_final))]
    return df_final, outliers

def throw_low_cov(df):
    filtered = df[df.freq > 1]
    return filtered

#input adapters for probabilities
def read_model_params(input_file, param_type):
   '''
   Converts user tsv input of custom model paramaters to dictionaries compatible with the
   original implementation
   Args:
      input_file (str): file name of input parameter tsv
      param_type (str): string designating which parameter dictionary to convert to, options: transitions, emissions, repeats, background
    Returns:
      param_dict (dictionary): dictionary of designated parameters
      
   '''
   #convert to respective dictionary
   if param_type in ["transitions","background"]: #input_df is a series here
    #convert to transition dictionary
    input_df = pd.read_csv(input_file, sep="\t")
    param_dict = input_df.iloc[0].to_dict()
    if param_type == 'background':
      param_dict[''] = 0
   if param_type in ["emissions","repeat"]: #input_df is a dataframe here
    #convert to transition dictionary
    input_df = pd.read_csv(input_file, sep="\t",index_col=0)
    if param_type == 'repeat':
       #add blank row
      print(input_df)
      input_df.loc[len( pd.DataFrame(input_df).index)] = 0
      input_df.index = ['A','G','C','T','']
    param_dict = input_df.to_dict()
    print(param_dict)
   return param_dict

def generate_input_sheet(coords_file, chrom_sizes_file, ref,flanking_length=200):
  '''
  Function to generate input sheet with the following info from 4 or 5 column bedfile with the following columns: chromosome, start, end, repeat motif, name
  (name is an optional column with default name being the coordinated in chr:start-end format)
  Args:
    coords_file (str): bed file with 4 or 5 columns in chromosome, start, end, repeat motif, name format
    chrom_sizes_file (str): chrom.sizes file containing the sizes of each chromosome in the reference, used to ensure prefix/suffix does not exceed the length of the chromosome
    ref (str): path to reference genome to use
    flanking_length (int): length of desired prefix and suffix, must be greater than the length of flank parameter (if given) or 30 (if default), recommended >100bp
  Returns:
    targets (DataFrame): pandas dataframe of targets with columns name, prefix, repeat, suffix
  '''
  #load reference genome
  ref_aligner = mappy.Aligner(ref)
  #load chromosome sizes
  chrom_sizes = pd.read_csv(chrom_sizes_file, sep="\t",names=["chr","chr_length"])
  #load bedfile
  coords=pd.read_csv(coords_file,sep="\t",header=None)
  #check if name column is included
  if len(coords.columns) == 5:
    coords.columns =  ["chr","start","end","repeat","name"]
  elif len(coords.columns) == 4:
    #add name column
    coords.columns = ["chr","start","end","repeat"]
    coords["name"] = coords.chr + ":" +coords.start.astype(str) + "-" + coords.end.astype(str)
  else:
    print("Bed file has unexpected number of columns, please ensure file has 4 (repeat motif) or 5 (motif and name) columns. Exiting...")
    return
  #get prefix and suffix coordinates
  coords['prefix_start'] = coords.start - flanking_length
  coords['suffix_end'] = coords.end + flanking_length
  #replace invalid flanking starts and ends
  coords_final = pd.merge(left=coords,right=chrom_sizes,how="left",on="chr")
  coords_final['prefix_start'] = np.where(coords_final['prefix_start'] < 0,0,coords_final['prefix_start'])
  coords_final['suffix_end'] = np.where(coords_final['suffix_end'] > coords_final['chr_length'], coords_final['chr_length'], coords_final['suffix_end'])

  #set up dictionary
  targets_dict = {"name":[],"prefix":[],"repeat":[],"suffix":[]}
  for target in coords_final.itertuples():
    targets_dict['name'].append(target.name)
    targets_dict["prefix"].append(ref_aligner.seq(name=target.chr,start=target.prefix_start,end=target.start))
    targets_dict['repeat'].append(target.repeat)
    targets_dict['suffix'].append(ref_aligner.seq(name=target.chr,start=target.end,end=target.suffix_end))
  return pd.DataFrame(targets_dict)


import pandas as pd
import mappy
from os.path import exists
from HMMSTR_utils.HMMSTR_utils import seq2int
from subprocess import run, PIPE
from importlib_resources import files

#custom imports
from process_vit_res.process_vit_res import *
class Process_Read:

    def __init__(self, header, seq, cutoff=30, mode = "map-ont", out = ".", k = None, w = None, use_full_seq = False):
        # self.read_id = header.split(" ")[0][1:]
        self.read_id = header #changed so we can just pass the read_id from different sources
        self.seq = seq
        self.cutoff = cutoff
        self.mode = mode
        self.use_full_seq = use_full_seq

        self.k = k
        self.w = w
        #check for if fastas exist?
        if exists(out + "_prefix.fa") and exists(out + "_suffix.fa"):
            self.prefix_fa = out + "_prefix.fa"
            self.suffix_fa = out+"_suffix.fa"
        elif exists(out + "_prefix.fa") == False:
            print("Prefix fasta file does not exist! Please check output path")
        else:
            print("Suffix fasta file does not exist! Please check output path")
        #call alignment and such to set other fields? <-- you can set fields in class methods without initializing them

    def align_mappy(self,target=None):
        '''
        This function aligns the prefixes and suffixes to the current read to assign it to the most likely target.
        '''
        if self.k  is None and self.w is None: # default
            aligner = mappy.Aligner(seq=self.seq,preset=self.mode,best_n = 1) #only return the best alignment per target
        #adjusted k
        elif self.k is not None and self.w is None:
            aligner = mappy.Aligner(seq=self.seq,preset=self.mode,best_n = 1, k =  self.k)
        elif self.k is None and self.w is not None:
            aligner = mappy.Aligner(seq=self.seq,preset=self.mode,best_n = 1, w = self.w)
        else:
            aligner = mappy.Aligner(seq=self.seq,preset=self.mode,best_n = 1, k =self.k , w= self.w) #custom parameters

        #align prefixes
        prefix_dict = {"name":[],"prefix_start":[],"prefix_end":[],"prefix_mapq":[],"strand":[],"alignment_length":[]}
        suffix_dict = {"name":[],"suffix_start":[],"suffix_end":[],"suffix_mapq":[],"strand":[],"alignment_length":[]}

        for name, seq,qual in mappy.fastx_read(self.prefix_fa):
            for hit in aligner.map(seq):
                if hit.mapq < self.cutoff:
                    continue
                #check if we are using bam file assignments, if we are only accept alignments to that target prefix 
                if target is not None:
                    if name != target:
                        continue
                prefix_dict["name"].append(name)
                prefix_dict["prefix_start"].append(hit.r_st)
                prefix_dict["prefix_end"].append(hit.r_en)
                prefix_dict["prefix_mapq"].append(hit.mapq)
                prefix_dict["strand"].append(hit.strand)
                prefix_dict["alignment_length"].append(hit.blen)

        #align_suffixes
        for name, seq,qual in mappy.fastx_read(self.suffix_fa):
            for hit in aligner.map(seq):
                if hit.mapq < self.cutoff:
                    continue
                if target is not None:
                    if name != target:
                        continue
                suffix_dict["name"].append(name)
                suffix_dict["suffix_start"].append(hit.r_st)
                suffix_dict["suffix_end"].append(hit.r_en)
                suffix_dict["suffix_mapq"].append(hit.mapq)
                suffix_dict["strand"].append(hit.strand)
                suffix_dict["alignment_length"].append(hit.blen)

        #record if no valid alignment found
        if len(prefix_dict['name']) < 1:
            self.prefix_df = False
        else:
            self.prefix_df = pd.DataFrame(prefix_dict)
        if len(suffix_dict['name']) < 1:
            self.suffix_df = False
        else:
            self.suffix_df = pd.DataFrame(suffix_dict)
        return
    
    def keep_region(self, prefix_info, suffix_info):
        '''
        This function determines if a read's alignments are eligible prefix and suffix alignments

        Parameters
        --------------------------------------------------------------------------------------------------
        prefix_info: pandas Series. Series contianing alignment info for the prefix
        suffix_info: pandas Series. Series containing alignment info for the suffix
        
        Returns
        --------------------------------------------------------------------------------------------------
        valid: boolean
            True if alignments suggest read is on target, False if alignments in the wrong orientation
        '''
        #check that both contain alignemnts
        if len(suffix_info.index) == 0 or len(prefix_info.index) == 0:
            return False
        #check alignments are on the same strand
        if prefix_info.strand[0] == suffix_info.strand[0]: 
            return True
        else:
            return False
        
    def get_align_info(self, row, prefix_info, suffix_info):
        '''
        This function gets attributes of a given read given that a target has been identified

        Parameters
        ----------------------------------------------------------------------------------------------------
        row: pandas Series. Row for current target.
        prefix_info: pandas Series. Series contianing alignment info for the prefix
        suffix_info: pandas Series. Series containing alignment info for the suffix

        Returns
        ----------------------------------------------------------------------------------------------------
        info: dictionary. Dictionary of alignment and subset information for the current read
        '''
        #dictionary of info for current target
        info = {}
        #may or may not want these in the dictionary, seems like a good idea to keep them together
        info["repeat"] = row.repeat.rstrip()
        info["prefix"] = row.prefix.rstrip()
        info["suffix"] = row.suffix.rstrip() 
        info["prefix_align_length"] = prefix_info.alignment_length[0]
        info["suffix_align_length"] = suffix_info.alignment_length[0]
        #get strand and start and end coordinates
        if prefix_info.strand[0] == 1 and suffix_info.strand[0] == 1:
            info["strand"] = "forward"
            info["align_start"] = prefix_info.prefix_start[0]
            info["align_end"] = suffix_info.suffix_end[0]
            info["end_length"] = info["suffix_align_length"]
            info["start_length"] = info["prefix_align_length"]
        else: #reverse
            info["strand"] = "reverse"
            info["align_start"] = suffix_info.suffix_start[0] #switched back cause I was wrong about which coordinates mappy returns for the reverse case
            info["align_end"] = prefix_info.prefix_end[0]
            info["end_length"] = info["prefix_align_length"]
            info["start_length"] = info["suffix_align_length"]
        #record mapqs
        info["prefix_mapq"] = prefix_info.prefix_mapq[0]
        info["suffix_mapq"] = suffix_info.suffix_mapq[0]

        if self.use_full_seq:
            info["subset"] = self.seq
            #ADDED record subset coords so we can calculate repeat coordinates
            info["subset_start"] = 0
            info["subset_end"] = len(self.seq) -1
            return info

        if info["align_start"] + info["start_length"] - 400 < 0 and info["align_end"]-info["end_length"]+400 < len(self.seq):
            info["subset"] = self.seq[info["align_start"] + info["start_length"] - 50: info["align_end"]-info["end_length"]+400]
            info["subset_start"] = info["align_start"] + info["start_length"] - 50
            info["subset_end"] = info["align_end"]-info["end_length"]+400 -1

        elif info["align_start"] + info["start_length"] - 400 > 0 and info["align_end"]-info["end_length"]+400 > len(self.seq):
            info["subset"] = self.seq[info["align_start"] + info["start_length"] - 400: info["align_end"]-info["end_length"]+50]
            info["subset_start"] = info["align_start"] + info["start_length"] - 400
            info["subset_end"] = info["align_end"]-info["end_length"]+50 -1
        elif info["align_start"] + info["start_length"] - 400 < 0 and info["align_end"]-info["end_length"]+400 > len(self.seq):
            if info["align_start"] + info["start_length"] - 50 < 0 and info["align_end"]-info["end_length"]+50 > len(self.seq):
                info["subset"] = self.seq[info["align_start"] : info["align_end"]]
                info["subset_start"] = info["align_start"]
                info["subset_end"] = info["align_end"]-1
            else:
                info["subset"] = self.seq[info["align_start"] + info["start_length"] - 50: info["align_end"]-info["end_length"]+50] #switch back to 50 if this doesnt help
                info["subset_start"] = info["align_start"] + info["start_length"] - 50
                info["subset_end"] = info["align_end"]-info["end_length"]+50 -1
        else:
            info["subset"] = self.seq[info["align_start"] + info["start_length"] - 400: info["align_end"]-info["end_length"]+400]
            info["subset_start"] = info["align_start"] + info["start_length"] - 400
            info["subset_end"] = info["align_end"]-info["end_length"]+400 -1
        if len(info["subset"]) < 1:
            return #return nothing if there is no repeat in this sequence, spurious alignment
        return info

    def assign_targets(self, targets_df):#TODO this would need to change to accomodate only having either suffix or prefix info, I currently filter reads on this
        '''
        This function takes alignment results for current process read object and determines which target(s)
        are in results

        Parameters
        -----------------------------------------------------------------------------------------------------
        targets_df: pandas DataFrame. DataFrame of tandem repeat target loci to compare to alignment results
        '''
        #check if any alignemnts returned
        if isinstance(self.prefix_df, (bool)) or isinstance(self.suffix_df, (bool)): #no alignments
            self.target_info = {}
            return False #need to decide on final returns for this function still
        #subset to only get targets that aligned according to mappy
        candidate_targets = targets_df[targets_df.name.isin(self.prefix_df.name)] #previously sub_targ
        #get the best alignments per target identified (previously sub_prefixes and sub_suffixes)
        candidate_prefix_aligns = self.prefix_df.groupby('name').head(1).reset_index()
        candidate_suffix_aligns = self.suffix_df.groupby('name').head(1).reset_index()
        self.target_info = {}
        #filter candidates that aren't in a compatible orientation and save final candidates
        for row in candidate_targets.itertuples():
            prefix_info =  candidate_prefix_aligns[candidate_prefix_aligns.name == row.name].reset_index()
            suffix_info = candidate_suffix_aligns[candidate_suffix_aligns.name == row.name].reset_index()
            #save valid regions' attributes
            if self.keep_region(prefix_info, suffix_info):
                self.target_info[row.name] = self.get_align_info(row, prefix_info, suffix_info)

    def run_viterbi(self,hmm_file,rev_hmm_file,hidden_states,rev_states,out,build_pre, prefix_idx,output_labelled_seqs):
        '''
        This function runs viterbi on the current read across all identified targets.

        Parameters
        ----------------------------------------------------------------------------------------------------
        hmm_file: str. Suffix for model file for Viterbi.
        rev_hmm_file: str. Suffix for reverse model file for Viterbi.
        hidden_states: str. Suffix for hidden_states file.
        rev_states: str. Suffix reverse hidden_states file.
        out: str. Output prefix
        build_pre: str. Prefix for all model files
        prefix_idx: int. Offset for prefix states
        output_labelled_seqs: bool. Output labelled read sequence

        Returns
        ----------------------------------------------------------------------------------------------------
        bool for if the run was successful
        '''
        #loop across all identified targets
        #if no targets, return
        if self.target_info == {} or self.target_info is None:
            return False
        for name in self.target_info.keys():
            #choose hmm to use based on strand of target
            curr_hmm_file = build_pre + "_" + name + hmm_file
            curr_rev_file = build_pre + "_" + name + rev_hmm_file
            if self.target_info[name]["strand"] == "forward":
                curr_hmm = curr_hmm_file
                curr_hidden_states_file = open(build_pre + "_" + name + hidden_states,'r')
                curr_states = curr_hidden_states_file.readline().split(".")
                curr_hidden_states_file.close()
            else:
                curr_hmm = curr_rev_file
                curr_hidden_states_rev_file = open(build_pre + "_" + name + rev_states,'r')
                curr_states = curr_hidden_states_rev_file.readline().split(".")
                curr_hidden_states_rev_file.close()
            #convert seqeunce to numeric so it is compatible with the C code
            numeric_seq = str(seq2int(self.target_info[name]["subset"].upper()))
            T = str(len(self.target_info[name]["subset"]))
            repeat_len = len(self.target_info[name]["repeat"])
            repeat_len_str = str(repeat_len)
            prefix_idx_str = str(3*prefix_idx + 1)
            test_hmm_cython_path = files('c_files').joinpath('test_hmm_cython.py')
            command = ['python',test_hmm_cython_path,curr_hmm,T,numeric_seq, repeat_len_str, prefix_idx_str]
            #run viterbi on current read
            result = run(command, universal_newlines=True,capture_output=True, text=True)
            vit_out = result.stdout

            #methods are either static in this class or will be imported by name
            labeled_seq, pointers,MLE = label_states(vit_out, curr_states)
            likelihood, sub_labels,repeats,context, final_repeat_like, repeat_start, repeat_end = calc_likelihood(vit_out, pointers,labeled_seq, curr_states, self.target_info[name]["subset"], self.target_info[name]["subset_start"],self.target_info[name]["subset_end"])
            #save state labels for KMeans method, if time we can figure out how to do this without saving a file
            label_file = open(out+"_"+ name + "_labeled_seqs.txt","a")
            label_file.write(self.read_id + "\t" +".".join(labeled_seq)+"\n")
            label_file.close()
            count = count_repeats(labeled_seq,pointers,repeat_len,self.target_info[name]["subset"])
            score = self.target_info[name]["prefix_mapq"] + self.target_info[name]["suffix_mapq"]

            out_file = open(out+"_"+name+"_counts.txt","a")
            out_file.write(self.read_id + " " + self.target_info[name]["strand"] + " "+ str(score) + " " + str(MLE) + " " + str(likelihood)+ " " + str(final_repeat_like) + " " + str(repeat_start) + " "+ str(repeat_end) + " "+ str(self.target_info[name]["align_start"]) + " "+str(self.target_info[name]["align_end"])+ " " + str(count) + "\n")
            out_file.close()
            #output labelled sequence to context file for given target if output_labelled_seqs is set
            if output_labelled_seqs:
                print_labelled(self.read_id,self.target_info[name]["strand"],sub_labels,context,pointers,out + "_labelled_seqs/"+name+"_context_labeled.txt")
        return True

        

            






        


#I dont need all of these, remove all unused at the end, many will be encapsulated by other files
#from collections import Counter, defaultdict
import argparse
import os, glob
import gzip
#import numpy as np
import pandas as pd
from time import perf_counter, sleep
import multiprocessing as mp
from multiprocessing import Pool, Queue, Process, Manager, Semaphore
import pickle as pkl
import sys
#custom imports
from profile_HMM.profile_HMM import ProfileHMM
from HMMSTR_utils.HMMSTR_utils import *
from process_read.process_read import Process_Read
from GMM_stats.GMM_stats import GMMStats
from KDE_stats.KDE_stats import KDE_cluster
import pysam
import mappy
import queue  # imported for using queue.Empty exception


class LoadFromFile (argparse.Action):
    #for reading from input file instead of command line
    def __call__ (self, parser, namespace, values, option_string = None):
        with values as f:
            # parse arguments in the file and store them in the target namespace
            parser.parse_args(f.read().split(), namespace)

def read_my_file(argv):
    '''
    Custom parser for file input
    '''
    new_argv = []
    with open(argv,"r") as params:
        for a in params: #read each file
            for b in a.rstrip().split(" "):
                new_argv.append(b)
    return new_argv
def write_input_command(args):
    '''
    Function to write out args to file compatible with read_my_file for future runs
    '''
    with open(args.out + "_run_input.txt","w") as outfile:
        outfile.write(vars(args)["subcommand"] + "\n")
        #check subcommand 
        if vars(args)["subcommand"] == "coordinates":
            #write out coordinates specific arguments
            outfile.write(vars(args)["coords_file"] + "\n")
            outfile.write(vars(args)["chrom_sizes_file"] + "\n")
            outfile.write(vars(args)["ref"] + "\n")
            outfile.write("--input_flank_length"+" "+str(vars(args)["input_flank_length"]) + "\n")
        else:
            #targets
            outfile.write(vars(args)["targets"] + "\n")
        outfile.write(vars(args)["out"] + "\n")
        outfile.write(vars(args)["inFile"] + "\n")

        #TODO supress bootstrapping options if bootstrapping isn't selected
        for key,val in vars(args).items():
            if key in ["subcommand","out","inFile","coords_file","targets","ref","chrom_sizes_file","input_flank_length"]: #these are positional fields, put these where they go first
                continue
            if val in [None, False]: #check if we want this parameter in our input
                continue
            if val == True: #stores true
                outfile.write("--"+key +"\n")
                continue
            outfile.write("--"+key + " "+ str(val)+"\n")
    return
def convert_to_fasta(tsv,out):
    '''
    Function to write "read" entry for a given target for both prefix and suffix
    fasta files to be used as queries in blastn

    Args:
        tsv (Pandas Dataframe): input tsv containing name, prefix, repeat, suffix
        out (str): prefix/path for outputted files

    Returns:
        nothing, writes prefix and suffix fasta files to out_prefix.fa and out_suffix.fa respectively

    '''
    for row in tsv.itertuples():
        curr_out = open(out+"_prefix.fa",'a')
        curr_out.write(">"+row.name+"\n")
        curr_out.write(row.prefix + "\n")
        curr_out.close()
        curr_out = open(out +"_suffix.fa",'a')
        curr_out.write(">"+row.name+"\n")
        curr_out.write(row.suffix + "\n")
        curr_out.close()
    return
def build_all(curr_target, out, background,alphabet,transitions,mismatch_probs, repeat_probs,flanking_size):
    '''
    Function to initialize a ProfileHMM object for a row in input targets tsv

    Parameters
    ----------------------------------------------------------------------------
    curr_target: pandas Series. Row from input tsv corresponding to a single target
    out: str. Output prefix
    background: dictionary. Dictionary of background probabilities to use as background frequencies
    alphabet: list. List of characters used to define alphabet of HMM
    transitions: dictionary. Dictionary of transition probabilites to encode in models
    mismatch_probs: dictionary. Dictionary of mismatch probabilites to encode as emission probabilites in models
    repeat_probs: dictionary. Dictionary of custom repeat match probabilities
    flanking_size: int. Number of bases to encode in both the prefix and suffix of each model
    Returns
    ------------------------------------------------------------------------------
    None, writes ProfileHMM in format needed for modified umdhmm-v1.02
    '''
    name = curr_target[0]
    prefix = curr_target[1].rstrip()
    repeat = curr_target[2].rstrip()
    suffix = curr_target[3].rstrip()
    #initialize ProfileHMM object
    curr_hmm = ProfileHMM(name=name,alphabet=alphabet, prefix=prefix, repeat=repeat, suffix=suffix, background=background, out=out, transitions=transitions,mismatch_probs=mismatch_probs, repeat_probs=repeat_probs,flanking_size=flanking_size)
    curr_hmm.build_profile("forward") #builds and writes forward ProfileHMM object
    curr_hmm.build_profile("reverse") #builds and writes reverse ProfileHMM object
    return

    
def fetch_reads(bam_file, regions, mapq_cutoff, *args):
    '''
    Function to fetch reads from a BAM file given a list of regions
    '''
    with pysam.AlignmentFile(bam_file, "rb") as bam:
        for region in regions:
            for read in bam.fetch(region[0],region[1], region[2]):
                if read.mapping_quality >= mapq_cutoff:
                    seq = mappy.revcomp(read.query_sequence) if read.is_reverse else read.query_sequence
                    read_id = read.query_name
                    #check if name for region is given, if it is, region is name given in regions tuple (4th element), if no 4th element, region is the string of the region
                    if len(region) > 3:
                        region_name = region[3]
                    else:
                        region_name = f'{region[0]}:{region[1]}-{region[2]}'
                    yield (read_id, seq, region_name) + args

def worker(args):
    # print('Worker function called')  # Add this line
    try:
        read_id, seq, region, hmm_file, rev_hmm_file, hidden_states, hidden_states_rev, out, targets, build_pre, mode, mapq_cutoff, k, w, use_full_read, flanking_size, output_labelled_seqs, coords_file = args
        # print('Calling process_read')  # Add this line
        process_read(read_id, seq, hmm_file, rev_hmm_file, hidden_states, hidden_states_rev, out, targets, build_pre, mode, mapq_cutoff, k, w, use_full_read, flanking_size, output_labelled_seqs, region, coords_file)
        return True
    except Exception as e:
        print(f'Exception in worker function: {e}')  # Add this line
def process_read(header,seq,hmm_file,rev_hmm_file,hidden_states,hidden_states_rev,out,targets, build_pre, mode, cutoff, k, w, use_full_seq, flanking_size,output_labelled_seqs,region=None, coord_file=None):
    '''
    Wrapper function to call all Process_Read methods. This will initialize all operations
    for a given read including (1) alignment (2) target identification and filtering 
    (3) repeat identification with viterbi (4) repeat counting and (5) recording relevant metrics

    Parameters
    ----------------------------------------------------------------------------------------------
    header: str. Header for read from fasta file
    seq: str. Raw read sequence
    hmm_file: str. Path to file containing written profile hmm from build_all
    rev_hmm_file: str. Path to file continaing the reverse strand profile hmm from build_all
    hidden_states: str. Path to hidden states file 
    hidden_states_rev: str. Path to hidden states file for reverse strand
    out: str. Output prefix for run including full or relative path
    targets: pandas DataFrame. DataFrame containing input information for each target to search for in reads
    build_pre: str. Path and prefix for hmm files, used in case targets have already been used and user doesn't want duplicate files
    mode: str. Sequencing technology used, map-ont for Nanopore, pb for PacBio
    cutoff: int. Mapq cutoff for mappy aligner to determine read maping location
    k: int. k-mer length parameter to be passed to mappy
    w: int. Minimizer window size to be passed to mappy
    use_full_seq: bool. Override to supress subsetting read sequence before Viterbi
    output_labelled_seqs: bool. Output labelled sequence output by Viterbi to a text file.
    region: str. Region this read mapped to, if known from bam file
    coord_file: str. Path to file containing coordinates for reads to be processed. Applicable in case of bam input.
    
    Returns
    ------------------------------------------------------------------------------------------------
    None, all outputs written to out
    '''
    #check if we are reading input from a bam file or a fastx file
    if coord_file is not None:
        read_id = header
        target = region
    elif coord_file is None:
        read_id = header.split(" ")[0][1:]
        target= None
    curr_read = Process_Read(read_id, seq, cutoff, mode, out, k,w, use_full_seq)
    curr_read.align_mappy(target) #sets prefix_df and suffix_df fields for read

    curr_read.assign_targets(targets) #sets target_info for current read, this is a dictionary of keys=target name and values=info, all information needed for the target
    #The check for no assigned targets is in run_vit
    targets_found = curr_read.run_viterbi(hmm_file=hmm_file,rev_hmm_file=rev_hmm_file,hidden_states=hidden_states, rev_states=hidden_states_rev, out=out,build_pre=build_pre, prefix_idx = flanking_size,output_labelled_seqs=output_labelled_seqs)
    return

def decide_method(target,out,out_count_name):
    '''
    Function to automatically choose a peakcalling method based on data IQR and outliers

    Parameters
    ----------------------------------------------------------------------------------------------
    target: pandas Series. Row from targets dataframe to generate method for
    out: str. Output prefix
    out_count_name: str. Path to file containing raw counts data for this target
    
    Returns
    ------------------------------------------------------------------------------------------------
    decision: str. Method to use for peakcalling
    '''
    #read in counts for current row
    name = target[0]
    out_count_file = out + "_" + name + out_count_name
    if os.path.exists(out_count_file):
        counts_file= pd.read_csv(out_count_file, sep=" ",header=None)
        counts_file.columns = ["read_id","strand","align_score","neg_log_likelihood","subset_likelihood","repeat_likelihood","repeat_start","repeat_end","align_start", "align_end","counts"]
        counts=counts_file.counts
    else:
        return None
    
    QR_range, IQR_range = get_IQR(counts,0.10)

    #number of outliers according to IQR
    lower_counts = len(counts[counts <= IQR_range[0]])
    upper_counts = len(counts[counts >= IQR_range[1]])
    #number of outliers according to quartiles -- not currently used, can be if tuning is necessary in the future
    lower_Q1_counts = len(counts[counts <= QR_range[0]])
    upper_Q3_counts = len(counts[counts <= QR_range[1]])

    data_width = max(counts) - min(counts)
    QR_width = QR_range[1]-QR_range[0]
    IQR_width = IQR_range[1] - IQR_range[0]

    #3 cases: (1) KDE--throw-outliers (2) KDE (3) GMM

    #check if we should use KDE, no outliers found -- can play with data_width definition here if need be
    if data_width < 4 or (QR_width <= 3 and data_width < 2*QR_width): #not including checking for outliers here rn, might need to
        decision = "kde"
    elif QR_width <= 4 and (lower_counts > 0 or upper_counts > 0): #might need to play with this
        decision = "kde_throw_outliers"
    else:
        decision = "gmm"
    return decision

def call_peaks(row, out, out_count_name, plot_hists, max_peaks, filter_outliers=False, filter_quantile=0.25,bootstrap=False, CI_width=0.95, resample_size=100,allele_specific_plots=False,allele_specif_CIs=False, bandwidth='scott',kernel="gaussian",flanking_like_filter=False):
    '''
    This method is the wrapper function for calling both KDE and GMM classes based on the decision per target, it also insures the outputs are uniform across all methods

    Parameters
    ----------------------------------------------------------------------------------------------
    row: pandas Series. Row from targets dataframe to call peaks for
    out: str. Output prefix
    out_count_name: str. Path to file containing raw counts data for this target
    plot_hists: bool. Boolean designating if plots should be saved.
    max_peaks: int. Maximum number of peaks to call.
    filter_outliers: bool. Boolean designating if outlier counts should be filtered before peak calling
    filter_quantile: float. If filter_outliers is set, determine outler counts based on this quantile.
    bootstrap: bool. Boolean designating if bootstrapping of allele calls should be performed.
    CI_width: float. If bootstrapping is set, calculate the confidence intervals at this width.
    resample_size: int. Number of iterations to resample for bootstraping.
    allele_specific_plots: bool. If true, output plots of allele count distributions in addition to overall read histograms and peak calls.
    allele_specif_CIs: bool. Output confidence intervals for each allele independently after clusters are assigned in addition to overall CIs.
    bandwidth: str or float. KDE specific parameter, either a string designating estimation method or float designating bandwidth to use.
    kernel: str. KDE specific parameter, KDE kernel to use in peakcalling
    flanking_like_filter: bool. If set, reads with exceedingly low likelihood at their flanking sequence will be filtered before clustering.
    
    Returns
    ------------------------------------------------------------------------------------------------
    final: pandas Series. Genotype calls and supporting information for final output.
    '''
    #make cluster object based on decision
    decision = row[4]

    if decision is not None and "kde" in decision:
        if decision == "kde_throw_outliers":
            filter_outliers = True #override when this option is chosen, will be the same as kde if user overrides
        curr_kde = KDE_cluster(row,out,out_count_name, filter_outliers,flanking_like_filter) #discard outliers will still be included in all of these regardless of if KDE-throw-outliers is chosen, that is the only option where it is the default
        if curr_kde.data is None: 
            allele_calls = {}
            allele_calls = {"name":curr_kde.name}
            allele_calls['bandwidth'] = -1
            allele_calls["peak_calling_method"] = -1
            allele_calls["num_supporting_reads"] = 0
            for i in range(max_peaks):
                allele_calls["A"+str(i + 1)+":median"] = 0
                allele_calls["A"+str(i + 1)+":mode"] = 0
                allele_calls["A"+str(i + 1)+":supporting_reads"] = 0
                allele_calls["A"+str(i+1)+":SD"] = 0
                if bootstrap:
                    allele_calls["A"+str(i)+":median_CI"] = (0,0)
                if allele_specif_CIs:
                    allele_calls["A"+ str(i+1)+":median_CI_allele_specific"] = (0,0)
            allele_calls_series = pd.Series(allele_calls)
            cols = allele_calls_series.index.tolist()
            cols.sort()
            final = allele_calls_series[cols]
            return final
        #get read info tsv
        curr_kde.data = curr_kde.get_stats(plot_hists, filter_outliers, filter_quantile, flanking_like_filter)
        
        clusters, allele_calls, outliers, flanking_outliers = curr_kde.call_clusters(kernel=kernel, bandwidth=bandwidth,max_k=max_peaks, output_plots =  plot_hists, filter_quantile=filter_quantile)
        assignments = curr_kde.assign_clusters(clusters,outliers, flanking_outliers)

        #plot allele_specific plots if applicable
        if allele_specific_plots:
            for assignment in assignments.cluster_assignments.unique():
                curr_kde.call_clusters(kernel=kernel, bandwidth=bandwidth,max_k=1, output_plots =  True, subset=assignments[assignments.cluster_assignments == assignment], allele_specific=True, allele=assignment, filter_quantile=filter_quantile)
    
        #write out read assignments
        assignments["name"] = curr_kde.name
        assignments["peak_calling_method"] = decision
        
        if flanking_like_filter:
            assignments[['name','read_id','strand', 'align_score','neg_log_likelihood', 'subset_likelihood', 'repeat_likelihood','repeat_start', 'repeat_end', 'align_start', 'align_end','counts', 'freq','cluster_assignments',"outlier","flanking_outlier","peak_calling_method"]].to_csv(out + "_read_assignments.tsv", sep="\t", header=None,index=False, mode="a")
        else:
            assignments[['name','read_id','strand', 'align_score','neg_log_likelihood', 'subset_likelihood', 'repeat_likelihood','repeat_start', 'repeat_end', 'align_start', 'align_end','counts', 'freq','cluster_assignments',"outlier","peak_calling_method"]].to_csv(out + "_read_assignments.tsv", sep="\t", index=False, mode="a", header=None)

        if len(assignments[assignments.outlier == False][assignments.flanking_outlier == False].cluster_assignments.unique()) < max_peaks:
            for i in range(max_peaks - len(assignments[assignments.outlier == False][assignments.flanking_outlier == False].cluster_assignments.unique())):
                j = len(assignments[assignments.outlier == False][assignments.flanking_outlier == False].cluster_assignments.unique())+1
                allele_calls["A"+str(i+j)+":median"] = 0
                allele_calls["A"+str(i+j)+":mode"] = 0
                allele_calls["A"+str(i+j)+":supporting_reads"] = 0
                allele_calls["A"+str(i+j)+":SD"] = 0
                if allele_specif_CIs:
                    allele_calls["A"+str(i+j)+":median_CI_allele_specific"] = (0,0)
        allele_calls["name"] = curr_kde.name
        allele_calls_series = pd.Series(allele_calls)
        #bootstrap
        assignments = pd.merge(left=assignments, right=curr_kde.data[["read_id","freq"]], how="left")
        if bootstrap:
            if clusters != -1:
                print("Starting bootstrap...")
                median_CIs = curr_kde.bootstrap_KDE(assignments[assignments.outlier == False][assignments.flanking_outlier == False], resample_size, CI_width, max_peaks, out)
                allele_calls_series = pd.concat([allele_calls_series,pd.Series(median_CIs)])
        if allele_specif_CIs: 
            for assignment in assignments.cluster_assignments.unique():
                if clusters == -1:
                    break
                if pd.isna(assignment):
                    continue
                print("Starting allele-specific bootstrap...")
                allele_calls_series["A"+ str(assignment)+":median_CI_allele_specific"] = curr_kde.bootstrap_KDE_allele_specific(assignments[assignments.cluster_assignments == assignment], resample_size, CI_width, out)
        
        #get total number of supporting reads
        allele_calls_series["num_supporting_reads"] = 0
        for k in range(max_peaks):
            allele_calls_series["num_supporting_reads"] = allele_calls_series["num_supporting_reads"] + allele_calls["A"+str(k+1) + ":supporting_reads"]
        allele_calls_series["peak_calling_method"] = decision
        cols = allele_calls_series.index.tolist()
        cols.sort()
        final = allele_calls_series[cols]
        return final
        
    else:#GMM is the only other option
        #check existence of count file, may not exist if no reads identified for a given target
        name = row[0]
        out_count_file = out + "_" + name + out_count_name
        #print(out_count_file)
        if os.path.exists(out_count_file) == False:
            print(out_count_file + " does not exist, writing null row for", name,"...")
            # returning null at the beginning of the run may be causing issues with apply, try returning a null row of the right dimensions
            curr_dict = {"name":name}
            curr_dict['bandwidth'] = -1
            curr_dict["peak_calling_method"] = -1
            for i in range(1,max_peaks+1):
                curr_median = "A"+str(i)+":median"
                curr_mode = "A"+str(i)+":mode"
                curr_sd = "A" + str(i) + ":SD"
                curr_support = "A" + str(i) + ":supporting_reads"
                curr_allele_spec = "A"+ str(i)+":median_CI_allele_specific"
                curr_boot = "A"+str(i)+":median_CI"
                if curr_median not in curr_dict.keys():
                    curr_dict[curr_median] = 0
                if curr_mode not in curr_dict.keys():
                    curr_dict[curr_mode] = 0
                if curr_sd not in curr_dict.keys():
                    curr_dict[curr_sd] = 0
                if curr_support not in curr_dict.keys():
                    curr_dict[curr_support] = 0
                if bootstrap and curr_boot not in curr_dict.keys():
                    curr_dict[curr_boot] = (0,0)
                if allele_specif_CIs and curr_allele_spec not in curr_dict.keys():
                    curr_dict[curr_allele_spec] = (0,0)
                curr_dict["num_supporting_reads"] = 0
            
            return pd.Series(curr_dict)
        #initialize GMMStats object
        gmm_stats = GMMStats(target_row=row) #contains all target attributes as well as E and A dictionaries
        final_data = gmm_stats.get_stats(out_count_file, out,plot_hists, filter_outliers, filter_quantile=filter_quantile,flanking_like_filter=flanking_like_filter)
        
        #peak calling
        final_data2 = final_data[final_data.outlier == False][final_data.flanking_outlier == False][final_data.counts != 0].copy()
        curr_row, final_data2['cluster_assignments'] = gmm_stats.call_peaks(final_data2, out, max_peaks, plot=plot_hists, save_allele_plots = allele_specific_plots)
        if curr_row is None:
            print("current row doesnt exist")
            return
        
        if bootstrap:
            print("Starting bootstrap...")
            curr_row = gmm_stats.bootstrap_gmm(curr_row,final_data2, resample_size, CI_width, max_peaks, out)
        if allele_specif_CIs:
            for assignment in final_data2['cluster_assignments'].unique():
                print("Starting allele-specific bootstrap...")
                median_CIs = gmm_stats.bootstrap_gmm_allele_specific(final_data2[final_data2.cluster_assignments == assignment],resample_size,CI_width,out)
                curr_row["A"+ str(assignment+1)+":median_CI_allele_specific"] = median_CIs
            #check if we have a value for every column in max_peaks
            if len(final_data2['cluster_assignments'].unique()) < max_peaks:
                for i in range(max_peaks - len(final_data2['cluster_assignments'].unique())):
                    j = len(final_data2['cluster_assignments'].unique())+1
                    curr_row["A"+str(i+j)+":median_CI_allele_specific"] = (0,0)

        #write out cluster assignments to file
        assignments = pd.merge(left = final_data2[['read_id','counts','cluster_assignments']],right=final_data, on = ["read_id","counts"], how="right")
        assignments["name"] = gmm_stats.name
        assignments["peak_calling_method"] = decision
        assignments["cluster_assignments"] = assignments["cluster_assignments"]+1

        if flanking_like_filter:
            assignments[['name','read_id','strand', 'align_score','neg_log_likelihood', 'subset_likelihood', 'repeat_likelihood','repeat_start', 'repeat_end', 'align_start', 'align_end','counts', 'freq','cluster_assignments',"outlier","flanking_outlier","peak_calling_method"]].to_csv(out + "_read_assignments.tsv", sep="\t", header=None,index=False, mode="a")
        else:
            assignments[['name','read_id','strand', 'align_score','neg_log_likelihood', 'subset_likelihood', 'repeat_likelihood','repeat_start', 'repeat_end', 'align_start', 'align_end','counts', 'freq','cluster_assignments',"outlier","peak_calling_method"]].to_csv(out + "_read_assignments.tsv", sep="\t", index=False, mode="a", header=None)
    #add bandwidth column for kde calls
    curr_row['bandwidth'] = -1
    curr_row["peak_calling_method"] = decision
    return curr_row #return genotype
    
def call_peaks_stranded(row, out, out_count_name, plot_hists, max_peaks, filter_outliers=False, filter_quantile=0.25,bootstrap=False, CI_width=0.95, resample_size=100,allele_specific_plots=False,allele_specif_CIs=False, bandwidth='scott',kernel="gaussian",flanking_like_filter=False, strand=None):
    '''
    This method is the wrapper function for calling both KDE and GMM classes based on the decision per target, it also insures the outputs are uniform across all methods. This version runs only on the given strand.

    Parameters
    ----------------------------------------------------------------------------------------------
    row: pandas Series. Row from targets dataframe to call peaks for
    out: str. Output prefix
    out_count_name: str. Path to file containing raw counts data for this target
    plot_hists: bool. Boolean designating if plots should be saved.
    max_peaks: int. Maximum number of peaks to call.
    filter_outliers: bool. Boolean designating if outlier counts should be filtered before peak calling
    filter_quantile: float. If filter_outliers is set, determine outler counts based on this quantile.
    bootstrap: bool. Boolean designating if bootstrapping of allele calls should be performed.
    CI_width: float. If bootstrapping is set, calculate the confidence intervals at this width.
    resample_size: int. Number of iterations to resample for bootstraping.
    allele_specific_plots: bool. If true, output plots of allele count distributions in addition to overall read histograms and peak calls.
    allele_specif_CIs: bool. Output confidence intervals for each allele independently after clusters are assigned in addition to overall CIs.
    bandwidth: str or float. KDE specific parameter, either a string designating estimation method or float designating bandwidth to use.
    kernel: str. KDE specific parameter, KDE kernel to use in peakcalling
    flanking_like_filter: bool. If set, reads with exceedingly low likelihood at their flanking sequence will be filtered before clustering.
    strand: str. Current strand to call peaks for.
    
    Returns
    ------------------------------------------------------------------------------------------------
    final: pandas Series. Genotype calls and supporting information for final output.
    '''
    #make cluster object based on decision
    decision = row[4]
    if decision is not None and "kde" in decision:
        if decision == "kde_throw_outliers":
            filter_outliers = True #override when this option is chosen, will be the same as kde if user overrides
        curr_kde = KDE_cluster(row,out,out_count_name, filter_outliers,flanking_like_filter,strand) #discard outliers will still be included in all of these regardless of if KDE-throw-outliers is chosen, that is the only option where it is the default
        if curr_kde.data is None:
            allele_calls = {}
            allele_calls = {"name":curr_kde.name}
            allele_calls['bandwidth'] = -1
            allele_calls["peak_calling_method"] = -1
            for i in range(max_peaks):
                allele_calls["A"+str(i + 1)+":median"] = 0
                allele_calls["A"+str(i + 1)+":mode"] = 0
                allele_calls["A"+str(i + 1)+":supporting_reads"] = 0
                allele_calls["A"+str(i+1)+":SD"] = 0
                if bootstrap:
                    allele_calls["A"+str(i)+":median_CI"] = (0,0)
                if allele_specif_CIs:
                    allele_calls["A"+ str(i+1)+":median_CI_allele_specific"] = (0,0)
            allele_calls["strand"] = strand
            allele_calls_series = pd.Series(allele_calls)
            cols = allele_calls_series.index.tolist()
            cols.sort()
            final = allele_calls_series[cols]
            return final #TODO figure out how to return in a stranded manner
        #get read info tsv
        curr_kde.data = curr_kde.get_stats(plot_hists, filter_outliers, filter_quantile, flanking_like_filter, strand=strand)
        
        clusters, allele_calls, outliers, flanking_outliers = curr_kde.call_clusters(kernel=kernel, bandwidth=bandwidth,max_k=max_peaks, output_plots =  plot_hists, filter_quantile=filter_quantile,strand=strand)
        assignments = curr_kde.assign_clusters(clusters,outliers, flanking_outliers)

        #plot allele_specific plots if applicable
        if allele_specific_plots:
            for assignment in assignments.cluster_assignments.unique():
                curr_kde.call_clusters(kernel=kernel, bandwidth=bandwidth,max_k=1, output_plots =  True, subset=assignments[assignments.cluster_assignments == assignment], allele_specific=True, allele=assignment, filter_quantile=filter_quantile, strand=strand)
    
        #write out read assignments
        assignments["name"] = curr_kde.name
        assignments["peak_calling_method"] = decision
        
        if flanking_like_filter:
            assignments[['name','read_id','strand', 'align_score','neg_log_likelihood', 'subset_likelihood', 'repeat_likelihood','repeat_start', 'repeat_end', 'align_start', 'align_end','counts', 'freq','cluster_assignments',"outlier","flanking_outlier","peak_calling_method"]].to_csv(out + "_"+strand+"_read_assignments.tsv", sep="\t", header=None,index=False, mode="a")
        else:
            assignments[['name','read_id','strand', 'align_score','neg_log_likelihood', 'subset_likelihood', 'repeat_likelihood','repeat_start', 'repeat_end', 'align_start', 'align_end','counts', 'freq','cluster_assignments',"outlier","peak_calling_method"]].to_csv(out + "_"+strand+"_read_assignments.tsv", sep="\t", index=False, mode="a", header=None)
        #account for differing column number for less than max_k peaks
        if len(assignments[assignments.outlier == False].cluster_assignments.unique()) < max_peaks:
            for i in range(max_peaks - len(assignments[assignments.outlier == False].cluster_assignments.unique())):
                allele_calls["A"+str(i+len(assignments[assignments.outlier == False].cluster_assignments.unique())+1)+":median"] = 0
                allele_calls["A"+str(i+len(assignments[assignments.outlier == False].cluster_assignments.unique())+1)+":mode"] = 0
                allele_calls["A"+str(i+len(assignments[assignments.outlier == False].cluster_assignments.unique())+1)+":supporting_reads"] = 0
                allele_calls["A"+str(i+len(assignments[assignments.outlier == False].cluster_assignments.unique())+1)+":SD"] = 0
                if allele_specif_CIs:
                    allele_calls["A"+str(i+len(assignments[assignments.outlier == False].cluster_assignments.unique())+1)+":median_CI_allele_specific"] = (0,0)
        allele_calls["name"] = curr_kde.name
        allele_calls["strand"] = strand
        allele_calls_series = pd.Series(allele_calls)

        #bootstrap
        assignments = pd.merge(left=assignments, right=curr_kde.data[["read_id","freq"]], how="left")
        if bootstrap:
            if clusters != -1:
                print("Starting bootstrap...")
                median_CIs = curr_kde.bootstrap_KDE(assignments[assignments.outlier == False], resample_size, CI_width, max_peaks, out)
                allele_calls_series = pd.concat([allele_calls_series,pd.Series(median_CIs)])
        if allele_specif_CIs: 
            for assignment in assignments.cluster_assignments.unique():
                if clusters == -1:
                    break
                if pd.isna(assignment):
                    continue
                print("Starting allele-specific bootstrap...")
                allele_calls_series["A"+ str(assignment)+":median_CI_allele_specific"] = curr_kde.bootstrap_KDE_allele_specific(assignments[assignments.cluster_assignments == assignment], resample_size, CI_width, out)
        
        #get total number of supporting reads
        allele_calls_series["num_supporting_reads"] = 0
        for k in range(max_peaks):
            allele_calls_series["num_supporting_reads"] = allele_calls_series["num_supporting_reads"] + allele_calls["A"+str(k+1) + ":supporting_reads"]
        allele_calls_series["peak_calling_method"] = decision
        cols = allele_calls_series.index.tolist()
        cols.sort()
        final = allele_calls_series[cols]
        return final 
    
    else:#GMM is the only other option
        #check existence of count file, may not exist if no reads identified for a given target
        name = row[0]
        out_count_file = out + "_" + name + out_count_name
        if os.path.exists(out_count_file) == False:
            print(out_count_file + " does not exist, writing null row for", name,"...")
            # returning null at the beginning of the run may be causing issues with apply, try returning a null row of the right dimensions
            curr_dict = {"name":name}
            curr_dict['bandwidth'] = -1
            curr_dict["peak_calling_method"] = -1
            for i in range(1,max_peaks+1):
                curr_median = "A"+str(i)+":median"
                curr_mode = "A"+str(i)+":mode"
                curr_sd = "A" + str(i) + ":SD"
                curr_support = "A" + str(i) + ":supporting_reads"
                #TODO add column names for alleles here, right now all of these are being set to (0,0) when they shouldn't be, somehow they are being reset to nan
                curr_allele_spec = "A"+ str(i)+":median_CI_allele_specific"
                curr_boot = "A"+str(i)+":median_CI"
                if curr_median not in curr_dict.keys():
                    curr_dict[curr_median] = 0
                if curr_mode not in curr_dict.keys():
                    curr_dict[curr_mode] = 0
                if curr_sd not in curr_dict.keys():
                    curr_dict[curr_sd] = 0
                if curr_support not in curr_dict.keys():
                    curr_dict[curr_support] = 0
                if bootstrap and curr_boot not in curr_dict.keys():
                    curr_dict[curr_boot] = (0,0)
                if allele_specif_CIs and curr_allele_spec not in curr_dict.keys():
                    curr_dict[curr_allele_spec] = (0,0)
                curr_dict["num_supporting_reads"] = 0
            curr_dict["strand"] = strand
            return pd.Series(curr_dict)
        #initialize GMMStats object
        gmm_stats = GMMStats(target_row=row) #contains all target attributes as well as E and A dictionaries
        final_data = gmm_stats.get_stats(out_count_file, out,plot_hists, filter_outliers, filter_quantile=filter_quantile,flanking_like_filter=flanking_like_filter, curr_strand=strand)
    
        final_data.to_csv(out +"_"+ gmm_stats.name +"_"+strand+"_final_out.tsv", index = False, sep="\t")
        
        #peak calling
        final_data2 = final_data[final_data.outlier == False][final_data.flanking_outlier == False][final_data.counts != 0].copy()
        curr_row, final_data2['cluster_assignments'] = gmm_stats.call_peaks(final_data2, out, max_peaks, plot=plot_hists, save_allele_plots = allele_specific_plots, strand=strand)
        if curr_row is None:
            print("current row doesnt exist")
            return
        
        if bootstrap:
            print("Starting bootstrap...")
            curr_row = gmm_stats.bootstrap_gmm(curr_row,final_data2, resample_size, CI_width, max_peaks, out)
        if allele_specif_CIs:
            for assignment in final_data2['cluster_assignments'].unique():
                median_CIs = gmm_stats.bootstrap_gmm_allele_specific(final_data2[final_data2.cluster_assignments == assignment],resample_size,CI_width,out)
                curr_row["A"+ str(assignment+1)+":median_CI_allele_specific"] = median_CIs
            if len(final_data2['cluster_assignments'].unique()) < max_peaks:
                for i in range(max_peaks - len(final_data2['cluster_assignments'].unique())):
                    j = len(final_data2['cluster_assignments'].unique())+1
                    curr_row["A"+str(i+j)+":median_CI_allele_specific"] = (0,0)
    
        #write out cluster assignments to file
        assignments = pd.merge(left = final_data2[['read_id','counts','cluster_assignments']],right=final_data, on = ["read_id","counts"], how="right")
        assignments["name"] = gmm_stats.name
        assignments["peak_calling_method"] = decision
        assignments["cluster_assignments"] = assignments["cluster_assignments"]+1
        if flanking_like_filter:
            assignments[['name','read_id','strand', 'align_score','neg_log_likelihood', 'subset_likelihood', 'repeat_likelihood','repeat_start', 'repeat_end', 'align_start', 'align_end','counts', 'freq','cluster_assignments',"outlier","flanking_outlier","peak_calling_method"]].to_csv(out + "_"+strand+"_read_assignments.tsv", sep="\t", header=None,index=False, mode="a")
        else:
            assignments[['name','read_id','strand', 'align_score','neg_log_likelihood', 'subset_likelihood', 'repeat_likelihood','repeat_start', 'repeat_end', 'align_start', 'align_end','counts', 'freq','cluster_assignments',"outlier","peak_calling_method"]].to_csv(out + "_"+strand+"_read_assignments.tsv", sep="\t", index=False, mode="a", header=None)
    #add bandwidth column for kde calls
    curr_row['bandwidth'] = -1
    curr_row["peak_calling_method"] = decision
    curr_row["strand"] = strand
    return curr_row


def main():
    parser = argparse.ArgumentParser()
    #Required inputs
    subparsers = parser.add_subparsers(dest='subcommand')
    subparsers.required = True
    #Target input tsv parser
    parser_infile = subparsers.add_parser('targets_tsv')
    parser_infile.add_argument(dest ="targets",type=str, help='TSV with name, prefix, repeat, suffix, header required. Recommended at least 100bp with default alignment parameters')
    #coordinates mode parser
    parser_coords = subparsers.add_parser('coordinates')
    parser_coords.add_argument(dest = 'coords_file',type=str,help='Path to input custom bed file with either 4 (chr,start,end,repeat motif) or 5 (chr,start,end,repeat motif, name) columns')
    parser_coords.add_argument(dest='chrom_sizes_file',type=str,help = 'Path to chromosome sizes file')
    parser_coords.add_argument(dest="ref",type=str,help="Path to reference genome to get flanking sequence from")
    parser_coords.add_argument("--input_flank_length",type=int,help="Length of prefix and suffix to get from reference, must be longer than 30bp (Default) or flanking_size parameter (default: %(default)s)",default=200)

    #same as above but for targets_tsv
    parser.add_argument(dest ="out",type=str, help='Output prefix including directory path')
    parser.add_argument(dest = "inFile", type=str, help= 'Sequence to search and annotate in fasta or fastq, gzipped accepted. Bam files accepted in coordinate mode.')

    #Optional output options
    parser.add_argument("--output_plots",help="Output plots, default outputs supporting read histograms and model of best fit plots per target", action='store_true')
    parser.add_argument("--output_labelled_seqs",help="Output the model path through prefix, repeat, and suffix identified per read as context_labelled.txt per target", action='store_true')
    parser.add_argument("--max_peaks", type=int,help="Maximum number of peaks to calculate AIC and BIC for in peak calling (default: %(default)s)",default=2)
    parser.add_argument("--cpus", type=int,help="Number of cpus to use. If none given, half of available cpus used (default: %(default)s)",default=int(mp.cpu_count()/2))
    parser.add_argument("--flanking_size", type=int,help="Number of basepairs flanking repeat to encode in model, longer recommended for highly repeatitive regions (default: %(default)s)",default=100)

    #alignment parameters
    parser.add_argument("--mode", type=str,help="map-ont (Nanopore), pb (PacBio), or sr (short accurate reads, use for accurate short flanking sequence input) (default: %(default)s)",default="map-ont")
    parser.add_argument("--mapq_cutoff", type=int,help="MapQ cutoff for prefix and suffix alignment (default: %(default)s)",default=30)
    parser.add_argument("--k", type=int,help="Kmer parameter to be passed to mappy",default=None)
    parser.add_argument("--w", type=int,help="Window parameter to be passed to mappy",default=None)
    parser.add_argument("--use_full_read", help="Flag to not subset the read by the alignment, optimal for reads with short prefix and suffix",action='store_true')

    #Peakcalling options
    parser.add_argument("--peakcalling_method", type=str, help="If set, this peak calling method will be used instead of chosen by decision process. Options: gmm, kde, or kde_throw_outliers (default: %(default)s)",default="auto")
    parser.add_argument("--bandwidth", type=int, help="Bandwidth to use for KDE (if chosen). Will default to scott method if not provided", default=None)
    parser.add_argument("--kernel", type=str, help="Kernel to use for KDE (if chosen) (default: %(default)s)", default="gaussian")

    #optional boostrapping parameters
    parser.add_argument("--bootstrap", help="Boolean designating to output boostrapped confidence intervals with genotype calls",action='store_true')
    parser.add_argument("--call_width", type=float, help="Decimal percentage designating width of confidence interval for allele calls (default: %(default)s)", default=0.95)
    parser.add_argument("--resample_size", type=int, help="Number of times to resample in bootstrapping (default: %(default)s)", default=100)

    #allele specific stats options
    parser.add_argument("--allele_specific_CIs",  help="Output allele-specific bootsrapped confidence intervals", action='store_true')
    parser.add_argument("--allele_specific_plots", help="Output allele-specific histograms with model of best fit", action='store_true')
    parser.add_argument("--discard_outliers", help="Discard outliers based on quantile", action='store_true')
    parser.add_argument("--filter_quantile", type=float, help="Float designating quantile of count frequency to discard when filtering outliers (default: %(default)s)", default=0.25)

    #post read processing filters/options
    parser.add_argument("--flanking_like_filter", help="If set, filter reads per target that are designated as outliers based on flanking sequence likelihood",action='store_true')
    parser.add_argument("--stranded_report", help="If set, genotypes are called for each strand separately as well as together and strand bias is reported if found",action='store_true')

    #for debugging
    parser.add_argument("--cluster_only", help="Run clustering on output corresponding to required arguments, can only be run after a run with --save_intermediates", action='store_true')
    parser.add_argument("--save_intermediates", help="Flag designating to save intermediate files such as model inputs, raw count files, and state sequence files", action='store_true')

    #optional, if input, models and accompanying files already produced and use the provided prefix
    parser.add_argument("--background", type=str, help="Pickle of custom background dictionary, must include empty string for deletion character and total must add up to 1", default=None)
    parser.add_argument("--E_probs", type=str, help= 'TSV of custom emissions probability matrix, see format specifications', default=None)
    parser.add_argument("--A_probs", type=str, help= 'TSV of custom transition probabilities, see format specifications', default=None)
    parser.add_argument("--custom_RM", type=str, help= 'TSV of custom repeat match state probability matrix, see format specifications, used for motif mosacism, single target only', default=None)
    parser.add_argument("--hmm_pre", type=str,help="Prefix for files produced by build function, use if running the same targets across multiple input files. Only compatible with --save_intermediates option from previous run")


    #check input to see if file or command line was used
    if len(sys.argv) != 2:
        args = parser.parse_args()
    elif len(sys.argv) == 2:
        if sys.argv[1] == "--help" or sys.argv[1] == "-h":
            args = parser.parse_args()
            sys.exit()
        args = parser.parse_args(read_my_file(sys.argv[1]))

    print("Parameters for this HMMSTR run: ")
    with open(args.out + "_run_parameters.txt","w") as outfile:
        for key,val in vars(args).items():
            print(key,":",val)
            outfile.write(key + ":"+ str(val)+"\n")
    write_input_command(args)

    if args.output_plots:
        directory = args.out + "_plots"
        if os.path.exists(directory) == False:
            os.mkdir(directory) 
            print("Directory '% s' created" % directory)
        else:
            print("Directory '% s' already exists! Plots will be output to existing directory..." % directory)
    if args.output_labelled_seqs:
        directory = args.out + "_labelled_seqs"
        if os.path.exists(directory) == False:
            os.mkdir(directory) 
            print("Directory '% s' created" % directory)
        else:
            print("Directory '% s' already exists! Labelled sequences will be output to existing directory..." % directory) #if running on the same targets as a previous run the labelled sequences will be appended to the current files
    #check input mode and check if inputs compatible for coordinates
    if args.subcommand == 'coordinates':
        #check inputs
        if args.input_flank_length < args.flanking_size:
            print("Prefix and suffix length (input_flank_length,%d) are incompatible with model flanking length (flanking_size, %d), exiting..."%(args.input_flank_length,args.flanking_size))
            return
        #generate infile from coordinates given
        print("Coordinates input selected, getting %dbp flanking sequence from reference at %s..."%(args.input_flank_length,args.ref))
        targets = generate_input_sheet(args.coords_file, args.chrom_sizes_file, args.ref, args.input_flank_length)
        print("Inputs generated from coordinates! Saving target inputs to %s..."%(args.out+"_inputs.tsv"))
        targets.to_csv(args.out + "_inputs.tsv", sep="\t",index=False)


    if os.path.exists(args.inFile) == False:
        print("Infile file does not exist, exiting...")
        return
    if args.subcommand == 'targets_tsv':
        #read in targets
        print("Target tsv input selected! Reading input from  %s..."%(args.targets))
        targets = pd.read_csv(args.targets, sep="\t")
    else:
        print("Targets generated from coordinates. Reading in target inputs from %s..."%(args.out + "_inputs.tsv"))
    #convert custom inputs
    if args.E_probs is not None:
        E_probs = read_model_params(args.E_probs,"emissions")
    else:
        E_probs = None
    if args.A_probs is not None:
        A_probs = read_model_params(args.A_probs,"transitions")
    else:
        A_probs = None
    if args.background is not None:
        background = read_model_params(args.background,"background")
    else:
        background = None
    if args.custom_RM is not None:
        custom_RM = read_model_params(args.custom_RM,"repeat")
        #print(custom_RM)
    else:
        custom_RM = None
    
    if args.bandwidth is None:
        args.bandwidth = 'scott'

    #default alphabet that is compatible with all methods
    alphabet = ['A','T','C','G','']
    if args.cluster_only == False:
        if args.hmm_pre is None:
            pool_start = perf_counter()
            #build all models
            targets.apply(build_all,axis=1,args=(args.out, background, alphabet,A_probs, E_probs, custom_RM, args.flanking_size))
            pool_end = perf_counter()
            print("All models built! finished .... time was: ", str(pool_end-pool_start))
            build_pre = args.out 
        else:
            print("Using previously written input files with prefix: " + args.hmm_pre)
            build_pre = args.hmm_pre
        convert_to_fasta(targets,args.out) #write out all target flanking sequences to fasta files to be used by mappy

        #file prefixes and suffixes to use
        hmm_file = ".hmm" #this is a suffix, does not include individual names
        rev_hmm_file = "_revcomp" + ".hmm"
        hidden_states = ".hidden_states.txt"
        hidden_states_rev = "_revcomp.hidden_states.txt"

        #start multiprocess of read processing
        pool_start = perf_counter()
        print("Starting multiprocess, cpu's in use: " + str(args.cpus))
        pool = mp.Pool(processes=args.cpus)

        #check input file type and pass to appropriate parser
        #bam file input
        if args.subcommand == 'coordinates' and args.inFile.endswith('.bam'): # check if bam infile
            with open(args.coords_file, 'r') as f:
                num_columns = len(f.readline().split())
            with open(args.coords_file, 'r') as f:
                if num_columns == 4:
                    regions = [(line.split()[0], int(line.split()[1]), int(line.split()[2])) for line in f] #FIXME this current implementation uses the coordinates from the bed file to get regions and ignores names (if provided), this will break in process read since we write our outputs in terms of names, thus currently only compatible with no names given
                elif num_columns == 5:
                    regions = [(line.split()[0], int(line.split()[1]), int(line.split()[2]),line.split()[4]) for line in f] #names are given and these will correspond to the outputs and model names
            pool.imap(worker, fetch_reads(args.inFile, regions, args.mapq_cutoff, hmm_file, rev_hmm_file, hidden_states, hidden_states_rev, args.out, targets, build_pre, args.mode, args.mapq_cutoff, args.k, args.w, args.use_full_read, args.flanking_size, args.output_labelled_seqs, args.coords_file), chunksize=10) #10 for now, ran faster than args.cpus
        elif args.subcommand == 'targets_tsv' and args.inFile.endswith('bam'): #not currently allowed
            print("Bam file input not allowed with targets_tsv input, please run with coordinates input or fast(a/q) infile. exiting...")
            return
        #check if compressed
        else:
            if args.inFile.endswith('gz'):
                #gzipped file
                if args.inFile.endswith("fasta.gz") or args.inFile.endswith("fa.gz"): #fasta file
                    [pool.apply_async(process_read, args=(header,seq,hmm_file,rev_hmm_file,hidden_states,hidden_states_rev,args.out,targets, build_pre, args.mode, args.mapq_cutoff, args.k, args.w, args.use_full_read, args.flanking_size, args.output_labelled_seqs)) for header, seq, bool in read_fasta(gzip.open(args.inFile,'rt'))]
                elif args.inFile.endswith("fastq.gz") or args.inFile.endswith("fq.gz"): #fastq
                    [pool.apply_async(process_read, args=(header,seq,hmm_file,rev_hmm_file,hidden_states,hidden_states_rev,args.out,targets, build_pre, args.mode, args.mapq_cutoff, args.k, args.w, args.use_full_read, args.flanking_size,args.output_labelled_seqs)) for header, seq, bool in read_fastq(gzip.open(args.inFile,'rt'))]
            else:
                if args.inFile.endswith('a'): #fasta file
                    [pool.apply_async(process_read, args=(header,seq,hmm_file,rev_hmm_file,hidden_states,hidden_states_rev,args.out,targets, build_pre, args.mode, args.mapq_cutoff, args.k, args.w, args.use_full_read,args.flanking_size,args.output_labelled_seqs)) for header, seq, bool in read_fasta(open(args.inFile))]
                elif args.inFile.endswith('q'): #fastq
                    [pool.apply_async(process_read, args=(header,seq,hmm_file,rev_hmm_file,hidden_states,hidden_states_rev,args.out,targets, build_pre, args.mode, args.mapq_cutoff, args.k, args.w, args.use_full_read,args.flanking_size,args.output_labelled_seqs)) for header, seq, bool in read_fastq(open(args.inFile))]
        pool.close()
        pool.join()
        pool_end = perf_counter()
        print("All reads processed, the pooled run took: ", pool_end-pool_start)

    #stats runs and allele calls
    if args.hmm_pre is None:
        build_pre = args.out #need to make sure this is compatible with new build all implentation
    else:
        build_pre = args.hmm_pre
    out_count_name = "_counts.txt"

    pool_start = perf_counter()

    #Choose a peakcalling method for allele calls
    if args.peakcalling_method == "auto":
        targets["peak_call_method"] = targets.apply(decide_method,args=(args.out, out_count_name), axis=1)
    elif args.peakcalling_method in ["kde","kde_throw_outliers","gmm"]:
        targets["peak_call_method"] = args.peakcalling_method
    else:
        print("Invalid method inputed for peak calling, using auto...")
        targets["peak_call_method"] = targets.apply(decide_method,args=(args.out, out_count_name), axis=1)

    #default output
    if args.stranded_report == False:
        #Note: this will override the current read_assignments file if it exists
        if args.flanking_like_filter:
            pd.DataFrame(['name','read_id','strand', 'align_score','neg_log_likelihood', 'subset_likelihood', 'repeat_likelihood','repeat_start', 'repeat_end', 'align_start', 'align_end','counts', 'freq','cluster_assignments',"outlier","flanking_outlier","peak_calling_method"]).T.to_csv(args.out + "_read_assignments.tsv", sep="\t", index=False, header=None)
        else:
            pd.DataFrame(['name','read_id','strand', 'align_score','neg_log_likelihood', 'subset_likelihood', 'repeat_likelihood','repeat_start', 'repeat_end', 'align_start', 'align_end', 'counts','freq','cluster_assignments',"outlier","peak_calling_method"]).T.to_csv(args.out + "_read_assignments.tsv", sep="\t", index=False,header=None)

        #call genotypes with assigned or given peakcalling method
        geno_df = targets.apply(call_peaks, args=(args.out, out_count_name, args.output_plots,args.max_peaks,args.discard_outliers,args.filter_quantile,args.bootstrap, args.call_width, args.resample_size,args.allele_specific_plots,args.allele_specific_CIs, args.bandwidth,args.kernel,args.flanking_like_filter), axis=1)
        #read resulting read_assignments for sorting
        read_assignments = pd.read_csv(args.out + "_read_assignments.tsv", sep="\t")
        #check if valid genotype output
        if isinstance(geno_df, pd.DataFrame) and isinstance(read_assignments, pd.DataFrame):
            #sort outputs
            geno_df_final = geno_df.apply(sort_outputs, args=(args.max_peaks,args.bootstrap,args.allele_specific_CIs), axis=1)
            read_assignments_final = sort_read_assign_alleles(read_assignments, geno_df_final, args.max_peaks)
            read_assignments_final.to_csv(args.out + "_read_assignments.tsv", sep="\t", index=False)
            geno_df_final.to_csv(args.out + "_genotype_calls.tsv",sep="\t",index=False)
            pool_end = perf_counter()
            print("Genotyping run done! Took: ", pool_end-pool_start)
        else:
            print(geno_df)
            print("Results are not a DataFrame! Something went wrong...")
    else: #stranded output
        if args.flanking_like_filter:
            pd.DataFrame(['name','read_id','strand', 'align_score','neg_log_likelihood', 'subset_likelihood', 'repeat_likelihood','repeat_start', 'repeat_end', 'align_start', 'align_end','counts', 'freq','cluster_assignments',"outlier","flanking_outlier","peak_calling_method"]).T.to_csv(args.out + "_forward_read_assignments.tsv", sep="\t", index=False,header=None)
            pd.DataFrame(['name','read_id','strand', 'align_score','neg_log_likelihood', 'subset_likelihood', 'repeat_likelihood','repeat_start', 'repeat_end', 'align_start', 'align_end','counts', 'freq','cluster_assignments',"outlier","flanking_outlier","peak_calling_method"]).T.to_csv(args.out + "_reverse_read_assignments.tsv", sep="\t", index=False,header=None)

        else:
            pd.DataFrame(['name','read_id','strand', 'align_score','neg_log_likelihood', 'subset_likelihood', 'repeat_likelihood','repeat_start', 'repeat_end', 'align_start', 'align_end', 'counts','freq','cluster_assignments',"outlier","peak_calling_method"]).T.to_csv(args.out + "_forward_read_assignments.tsv", sep="\t", index=False,header=None)
            pd.DataFrame(['name','read_id','strand', 'align_score','neg_log_likelihood', 'subset_likelihood', 'repeat_likelihood','repeat_start', 'repeat_end', 'align_start', 'align_end', 'counts','freq','cluster_assignments',"outlier","peak_calling_method"]).T.to_csv(args.out + "_reverse_read_assignments.tsv", sep="\t", index=False,header=None)

        #genotype each strand independently
        geno_forward_df = targets.apply(call_peaks_stranded, args=(args.out, out_count_name, args.output_plots,args.max_peaks,args.discard_outliers,args.filter_quantile,args.bootstrap, args.call_width, args.resample_size,args.allele_specific_plots,args.allele_specific_CIs, args.bandwidth,args.kernel,args.flanking_like_filter,"forward"), axis=1)
        geno_reverse_df = targets.apply(call_peaks_stranded, args=(args.out, out_count_name, args.output_plots,args.max_peaks,args.discard_outliers,args.filter_quantile,args.bootstrap, args.call_width, args.resample_size,args.allele_specific_plots,args.allele_specific_CIs, args.bandwidth,args.kernel,args.flanking_like_filter, "reverse"), axis=1)
        
        #concat stranded results
        geno_df = pd.concat([geno_forward_df,geno_reverse_df]).sort_values(by="name")
        if isinstance(geno_df, pd.DataFrame):
            geno_df_final = geno_df.apply(sort_outputs, args=(args.max_peaks,args.bootstrap,args.allele_specific_CIs,True), axis=1)
            for i in range(1,args.max_peaks+1):
                geno_df_final["A"+str(i)+":strand_std"] = geno_df_final.groupby("name")["A"+str(i)+":median"].transform(np.std)
            geno_df_final.to_csv(args.out + "_genotype_calls.tsv",sep="\t",index=False)
            pool_end = perf_counter()
            print("Genotyping done! Took: ", pool_end-pool_start)
        else:
            print(geno_df)
            print("Results are not a DataFrame! Something went wrong...")
    #clean up intermediates
    if not args.save_intermediates:
        for file_subset in glob.glob(args.out+"*hidden_states.txt"):
            if os.path.exists(file_subset):
                os.remove(file_subset)
        for file_subset in glob.glob(args.out+"*.hmm"):
            if os.path.exists(file_subset):
                os.remove(file_subset)
        for file_subset in glob.glob(args.out+"*labeled_seqs.txt"):
            if os.path.exists(file_subset):
                os.remove(file_subset)
        #FIXME counts file is required for cluster only
        for file_subset in glob.glob(args.out+"*counts.txt"):
            os.remove(file_subset)
        if os.path.exists(args.out+"_prefix.fa"):
            os.remove(args.out+"_prefix.fa")
        if os.path.exists(args.out+"_suffix.fa"):
            os.remove(args.out+"_suffix.fa")
if __name__ == "__main__":
  main()